import pya
from pya import *
from SiEPIC.utils import get_technology_by_name

class Waveguide_Arc(pya.PCellDeclarationHelper):
  """
  The PCell declaration for the waveguide arc.
  
  Author: Mustafa Hammood     mustafa@siepic.com
  """

  def __init__(self):

    # Important: initialize the super class
    super(Waveguide_Arc, self).__init__()
    TECHNOLOGY = get_technology_by_name('EBeam')

    # declare the parameters
    self.param("silayer", self.TypeLayer, "Si Layer", default = TECHNOLOGY['Si'])
    self.param("radius", self.TypeDouble, "Radius", default = 10)
    self.param("wg_width", self.TypeDouble, "Waveguide Width", default = 0.5)
    self.param("start_angle", self.TypeDouble, "Start Angle", default = 0)
    self.param("stop_angle", self.TypeDouble, "Stop Angle", default = 45)
    self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
    self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])
    # hidden parameters, can be used to query this component:
    self.param("p1", self.TypeShape, "DPoint location of pin1", default = Point(-10000, 0), hidden = True, readonly = True)
    self.param("p2", self.TypeShape, "DPoint location of pin2", default = Point(0, 10000), hidden = True, readonly = True)
    

  def display_text_impl(self):
    # Provide a descriptive text for the cell
    return "Waveguide_Arc(R=" + ('%.3f' % self.radius) + ")"

  def produce(self, layout, layers, parameters, cell):
    """
    coerce parameters (make consistent)
    """
    self._layers = layers
    self.cell = cell
    self._param_values = parameters
    self.layout = layout


    # cell: layout cell to place the layout
    # LayerSiN: which layer to use
    # r: radius
    # w: waveguide width
    # start_angle: starting angle of the arc
    # stop_agnle: stopping angle of the arc
    # length units in dbu
    import math
    from math import pi, cos, sin
    from SiEPIC.utils import arc_wg
    
    # fetch the parameters
    dbu = self.layout.dbu
    ly = self.layout
    
    LayerSi = self.silayer
    LayerSiN = self.silayer_layer
#    LayerSiN = ly.layer(LayerSi)
    LayerPinRecN = ly.layer(self.pinrec)
    LayerDevRecN = ly.layer(self.devrec)
    
    from SiEPIC.extend import to_itype
    w = to_itype(self.wg_width,dbu)
    r =  to_itype(self.radius,dbu)
    start_angle = self.start_angle
    stop_angle = self.stop_angle
    
    if start_angle > stop_angle:
      start_angle = self.stop_angle
      stop_angle = self.start_angle
    
    deg_to_rad = math.pi / 180.0
    
    
    # draw the arc
    x = 0
    y = 0
    
    
    from SiEPIC._globals import PIN_LENGTH as pin_length
    
    self.cell.shapes(LayerSiN).insert(arc_wg(r, w, start_angle, stop_angle))
    # Create the pins, as short paths:

    # Pin on the right side:
    x = r*math.cos( start_angle * deg_to_rad ) 
    y = r*math.sin( start_angle * deg_to_rad )
    
    x_pin = math.cos( (90 - start_angle)*deg_to_rad ) *pin_length/2
    y_pin = math.sin( (90 - start_angle)*deg_to_rad ) *pin_length/2
    
    p2 = [Point(x-x_pin, y+y_pin), Point(x+x_pin, y-y_pin)]
    p2c = Point(x, y)
    self.set_p2 = p2c
    self.p2 = p2c
    pin = Path(p2, w)
    self.cell.shapes(LayerPinRecN).insert(pin)
    t = Trans(Trans.R0, x, y)
    text = Text ("pin2", t)
    shape = self.cell.shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu


    # Pin on the left side:
    x = round(r*math.cos( stop_angle * deg_to_rad ))
    y = round(r*math.sin( stop_angle * deg_to_rad ))
    
    x_pin = math.cos( (90.0 - stop_angle)*deg_to_rad ) *pin_length/2
    y_pin = math.sin( (90.0 - stop_angle)*deg_to_rad ) *pin_length/2
    
    p1 = [Point(x+x_pin, y-y_pin), Point(x-x_pin, y+y_pin)]
    p1c = Point(x,y)
    self.set_p1 = p1c
    self.p1 = p1c
    pin = Path(p1, w)
    self.cell.shapes(LayerPinRecN).insert(pin)
    t = Trans(Trans.R0, x, y)
    text = Text ("pin1", t)
    shape = self.cell.shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu

    
    # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
    x = 0
    y = 0
    #layout_arc_wg_dbu(self.cell, LayerDevRecN, x, y, r, w*3, start_angle, stop_angle)
    self.cell.shapes(LayerDevRecN).insert(arc_wg(r, w*3, start_angle, stop_angle))



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name

class ebeam_dc(pya.PCellDeclarationHelper):
  """
  The PCell declaration for the Directional Coupler
  by Lukas Chrostowski, 2018/09
  compact model in INTERCONNECT based on:
   - https://kx.lumerical.com/t/lcml-directional-coupler-based-on-lookup-table-lcml-dc-strip-1550-lookuptable/2094
   - only parameterized for Lc, for the 500 x 220 nm waveguide with 200 nm gap, and 5 micron radius
   could be improved:
   - https://kx.lumerical.com/t/lcml-directional-coupler-based-on-analytical-functions-lcml-dc-strip-1550-analytical/2091
    
  """

  def __init__(self):

    # Important: initialize the super class
    super(ebeam_dc, self).__init__()
    TECHNOLOGY = get_technology_by_name('EBeam')

    # declare the parameters
    self.param("Lc", self.TypeDouble, "Coupler Length", default = 10.0)
    self.param("r", self.TypeDouble, "Radius", default = 10)
    self.param("w", self.TypeDouble, "Waveguide Width", default = 0.5)
    self.param("g", self.TypeDouble, "Gap", default = 0.2)
    self.param("silayer", self.TypeLayer, "Si Layer", default = TECHNOLOGY['Si'])
    self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
    self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])
    self.param("textl", self.TypeLayer, "Text Layer", default = LayerInfo(10, 0))

  def display_text_impl(self):
    # Provide a descriptive text for the cell
    return "ebeam_dc(Lc=" + ('%.3f' % self.Lc) + ")"

  def can_create_from_shape_impl(self):
    return False
    
  def produce_impl(self):
    # This is the main part of the implementation: create the layout

    # Fixed PCell parameters
    port_spacing = 2000  # spacing of the two ports, determines the angle required by the s-bend.

    from math import pi, cos, sin, acos
    from SiEPIC.utils import arc_wg, arc_wg_xy
    from SiEPIC._globals import PIN_LENGTH

    # fetch the parameters
    dbu = self.layout.dbu
    ly = self.layout
    shapes = self.cell.shapes
    LayerSiN = ly.layer(self.silayer)
    LayerPinRecN = ly.layer(self.pinrec)
    LayerDevRecN = ly.layer(self.devrec)
    TextLayerN = ly.layer(self.textl)

    Lc = int(round(self.Lc/dbu))
    w = int(round(self.w/dbu))
    r = int(round(self.r/dbu))
    g = int(round(self.g/dbu))
            
    # Create the parallel waveguides
    if Lc > 0:
      wg1 = Box(-Lc/2, -w/2+(w+g)/2, Lc/2, w/2+(w+g)/2)
      shapes(LayerSiN).insert(wg1)
      wg1 = Box(-Lc/2, -w/2-(w+g)/2, Lc/2, w/2-(w+g)/2)
      shapes(LayerSiN).insert(wg1)
      
    dc_angle = acos((r-abs(port_spacing/2))/r)*180/pi

    # bottom S-bends
    self.cell.shapes(LayerSiN).insert(arc_wg_xy(Lc/2,-r-(w+g)/2, r, w, 90-dc_angle, 90))
    self.cell.shapes(LayerSiN).insert(arc_wg_xy(-Lc/2, -r-(w+g)/2, r, w, 90, 90+dc_angle))
    y_bottom = round(-2*(1-cos(dc_angle/180.0*pi))*r)-(w+g)/2
    x_bottom = round(2*sin(dc_angle/180.0*pi)*r)
    t = Trans(Trans.R0,-x_bottom-Lc/2, y_bottom+r)
    self.cell.shapes(LayerSiN).insert(arc_wg(r, w, -90, -90+dc_angle).transformed(t))
    t = Trans(Trans.R0,x_bottom+Lc/2, y_bottom+r)
    self.cell.shapes(LayerSiN).insert(arc_wg(r, w, -90-dc_angle, -90).transformed(t))

    # top S-bends
    self.cell.shapes(LayerSiN).insert(arc_wg_xy(Lc/2,r+(w+g)/2, r, w, 270, 270+dc_angle))
    self.cell.shapes(LayerSiN).insert(arc_wg_xy(-Lc/2, r+(w+g)/2, r, w, 270-dc_angle, 270))
    y_top = round(2*(1-cos(dc_angle/180.0*pi))*r)+(w+g)/2
    x_top = round(2*sin(dc_angle/180.0*pi)*r)
    t = Trans(Trans.R0,-x_top-Lc/2, y_top-r)
    self.cell.shapes(LayerSiN).insert(arc_wg(r, w, 90-dc_angle, 90).transformed(t))
    t = Trans(Trans.R0,x_top+Lc/2, y_top-r)
    self.cell.shapes(LayerSiN).insert(arc_wg(r, w, 90, 90+dc_angle).transformed(t))
    
    # Pins on the bottom waveguide side:
    pin = Path([Point(-x_bottom+PIN_LENGTH/2-Lc/2, y_bottom), Point(-x_bottom-PIN_LENGTH/2-Lc/2, y_bottom)], w)
    shapes(LayerPinRecN).insert(pin)
    text = Text ("pin1", Trans(Trans.R0, -x_bottom-Lc/2, y_bottom))
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu

    pin = Path([Point(x_bottom-PIN_LENGTH/2+Lc/2, y_bottom), Point(x_bottom+PIN_LENGTH/2+Lc/2, y_bottom)], w)
    shapes(LayerPinRecN).insert(pin)
    text = Text ("pin3", Trans(Trans.R0, x_bottom+Lc/2, y_bottom))
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu

    # Pins on the top waveguide side:
    pin = Path([Point(-x_bottom+PIN_LENGTH/2-Lc/2, y_top), Point(-x_bottom-PIN_LENGTH/2-Lc/2, y_top)], w)
    shapes(LayerPinRecN).insert(pin)
    text = Text ("pin2", Trans(Trans.R0, -x_bottom-Lc/2, y_top))
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu

    pin = Path([Point(x_bottom-PIN_LENGTH/2+Lc/2, y_top), Point(x_bottom+PIN_LENGTH/2+Lc/2, y_top)], w)
    shapes(LayerPinRecN).insert(pin)
    text = Text ("pin4", Trans(Trans.R0, x_bottom+Lc/2, y_top))
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu

    # Merge all the waveguide shapes, to avoid any small gaps
    layer_temp = self.layout.layer(LayerInfo(913, 0))
    shapes_temp = self.cell.shapes(layer_temp)
    ShapeProcessor().merge(self.layout,self.cell,LayerSiN,shapes_temp,True,0,True,True)
    self.cell.shapes(LayerSiN).clear()
    shapes_SiN = self.cell.shapes(LayerSiN)
    ShapeProcessor().merge(self.layout,self.cell,layer_temp, shapes_SiN,True,0,True,True)
    self.cell.shapes(layer_temp).clear()

    # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
    dev = Box(-x_bottom-Lc/2, y_bottom-w/2-w, x_bottom+Lc/2, y_top+w/2+w )
    shapes(LayerDevRecN).insert(dev)

    # Compact model information
    t = Trans(Trans.R0, 0, -w)
    text = Text ("Lumerical_INTERCONNECT_library=Design kits/ebeam_v1.2", t)
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = r*0.017
    t = Trans(Trans.R0, 0, 0)
    text = Text ('Component=NO_MODEL', t)
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = r*0.017
    t = Trans(Trans.R0, 0, w)
    text = Text ('Spice_param:wg_width=%.3fu gap=%.3fu radius=%.3fu Lc=%.3fu'%(w*dbu,g*dbu,r*dbu,self.Lc), t)
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = r*0.017

    # print("Done drawing the layout for - ebeam_dc: %.3f" % ( self.Lc) )


import pya
from pya import *
from SiEPIC.utils import get_technology_by_name

class Contra_DC_Bent(pya.PCellDeclarationHelper):
  """
  The PCell declaration for the bent-CDC half ring resonator pcell
  
  Author: Mustafa Hammood     mustafa@siepic.com
  """

  def __init__(self):

    # Important: initialize the super class
    super(Contra_DC_Bent, self).__init__()
    TECHNOLOGY = get_technology_by_name('EBeam')

    # declare the parameters
    self.param("silayer", self.TypeLayer, "Si Layer", default = TECHNOLOGY['Si'])
    self.param("silayer_gratings", self.TypeLayer, "Si Gratings Layer", default = TECHNOLOGY['Si'])
    self.param("radius", self.TypeDouble, "Radius (um)", default = 25)
    self.param("gap", self.TypeDouble, "Gap (um)", default = 0.28)
    self.param("bus_width", self.TypeDouble, "Bus Width (um)", default = 0.45)
    self.param("ring_width", self.TypeDouble, "Ring Width (um)", default = 0.55)
    
    self.param("period", self.TypeDouble, "Gratings Period (nm)", default = 318)
    self.param("deltaWB", self.TypeDouble, "Bus Corrugation Width (um)", default = 0.04)
    self.param("deltaWR", self.TypeDouble, "Ring Corrugation Width (um)", default = 0.05)
    self.param("gamma", self.TypeDouble, "N (number of corrugations)", default = 135)
    
    self.param("busBend", self.TypeDouble, "Bus-to-straight bend raidus", default = 5)
    
    self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
    self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])
    # hidden parameters, can be used to query this component:
    self.param("p1", self.TypeShape, "DPoint location of pin1", default = Point(-10000, 0), hidden = True, readonly = True)
    self.param("p2", self.TypeShape, "DPoint location of pin2", default = Point(0, 10000), hidden = True, readonly = True)
    

  def display_text_impl(self):
    # Provide a descriptive text for the cell
    return "Contra_DC_Bent_(R=" + ('%.3f' % self.radius) + ")"

  def produce(self, layout, layers, parameters, cell):
    """
    coerce parameters (make consistent)
    """
    self._layers = layers
    self.cell = cell
    self._param_values = parameters
    self.layout = layout


    # cell: layout cell to place the layout
    # LayerSiN: which layer to use
    # r: radius
    # w: waveguide width
    # length units in dbu

    from math import pi, cos, sin
    from SiEPIC.utils import arc_wg, arc_wg_xy
    from SiEPIC._globals import PIN_LENGTH
    
    # fetch the parameters
    dbu = self.layout.dbu
    ly = self.layout
    
    LayerSi = self.silayer
    LayerSiN_gratings = self.silayer_gratings_layer
    LayerSiN = self.silayer_layer
    LayerPinRecN = ly.layer(self.pinrec)
    LayerDevRecN = ly.layer(self.devrec)
    
    from SiEPIC.extend import to_itype
    
    w_bus = to_itype(self.bus_width,dbu)
    w_ring = to_itype(self.ring_width,dbu)
    w = w_ring
    r = int(round(self.radius/dbu))
    gap = int(round(self.gap/dbu))
    period = self.period
    deltaWR = int(round(self.deltaWR/dbu))
    N = int(self.gamma)
    deltaWB = int(round(self.deltaWB/dbu))
    busBendR = int(round(self.busBend/dbu))

    
    # Center of everything
    x = 0
    y = 0

    # Angle of CDC portion, also bend angle!
    periodAngle = (180/pi) * (period/2) / (r + w_ring/2 + gap/2) 
    
    # Bend angle
    bendAngle = (180/pi) * (N*period/2) / (r + w_ring/2 + gap/2) 
    
    # Bend radius of bus
    rBus = r + gap + w_bus/2 +w_ring/2
      
    N_input = N;
    # Normalized number of corrugations to periodAngle
    N = N*periodAngle*2;

    # Convention:   Set A is the CDC portion that is not offsetted, set B is offsetted by corrugation width
      
    # Ring CDCs (set A)
    ii = periodAngle*2
    while ii < N+periodAngle*1.5:
      self.cell.shapes(LayerSiN_gratings).insert(arc_wg_xy(x,y, r-deltaWR/2, w_ring, 90+bendAngle-ii, 90+bendAngle-ii-periodAngle))
      ii = ii+periodAngle
      ii = ii+periodAngle
      
    # Ring CDCs (set B)
    ii = periodAngle
    while ii < N:
      self.cell.shapes(LayerSiN_gratings).insert(arc_wg_xy(x,y, r+deltaWR/2, w_ring, 90+bendAngle-ii, 90+bendAngle-ii-periodAngle))
      ii = ii+periodAngle
      ii = ii+periodAngle


    # Bus CDCs (set A)
    ii = periodAngle
    while ii < N:
      self.cell.shapes(LayerSiN_gratings).insert(arc_wg_xy(x,y, rBus-deltaWB/2, w_bus, 90+bendAngle-ii, 90+bendAngle-ii-periodAngle))
      ii = ii+periodAngle
      ii = ii+periodAngle

    # Bus CDCs (set B)
    ii = periodAngle*2
    while ii < N+periodAngle*1.5:
      self.cell.shapes(LayerSiN_gratings).insert(arc_wg_xy(x,y, rBus+deltaWB/2, w_bus, 90+bendAngle-ii, 90+bendAngle-ii-periodAngle))
      ii = ii+periodAngle
      ii = ii+periodAngle

    # Ring non-CDC left
    self.cell.shapes(LayerSiN).insert(arc_wg_xy(x,y, r, w_ring, 180-(90-bendAngle), 180))
    
    # Ring non-CDC right
    self.cell.shapes(LayerSiN).insert(arc_wg_xy(x,y, r, w_ring, 0, 90-bendAngle))
    
    bendAngleRad = (pi/180) * bendAngle
    # Draw Bus non-CDC Waveguide
    y_center = rBus * cos(bendAngleRad) + busBendR * cos(bendAngleRad);
    x_center = rBus * sin(bendAngleRad) + busBendR * sin(bendAngleRad);
    
    self.cell.shapes(LayerSiN).insert(arc_wg_xy(x_center,y_center, busBendR, w_bus, 270-bendAngle,270))
    self.cell.shapes(LayerSiN).insert(arc_wg_xy(-x_center,y_center, busBendR, w_bus, 270, 270-bendAngle))
    
    # Create the pins, as short paths:
    from SiEPIC._globals import PIN_LENGTH as pin_length
    
    # Pin on the right side:
    y_center = rBus * cos(bendAngleRad) + busBendR * cos(bendAngleRad) - busBendR;
    p2 = [Point(-pin_length/2 +x_center, y_center), Point(pin_length/2 +x_center, y_center)]
    p2c = Point(x_center, y_center)
    self.set_p2 = p2c
    self.p2 = p2c
    pin = Path(p2, w_bus)
    self.cell.shapes(LayerPinRecN).insert(pin)
    t = Trans(Trans.R0, x_center, y_center)
    text = Text ("pin2", t)
    shape = self.cell.shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu
    
    # Pin on the left side:
    x_center = -x_center
    p1 = [Point(pin_length/2 +x_center, y_center), Point(-pin_length/2 +x_center, y_center)]
    p1c = Point(x_center, y_center)
    self.set_p1 = p1c
    self.p1 = p1c
    pin = Path(p1, w_bus)
    self.cell.shapes(LayerPinRecN).insert(pin)
    t = Trans(Trans.R0, x_center, y_center)
    text = Text ("pin1", t)
    shape = self.cell.shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu
    
    
    # Pin on the bottom left side:
    p3 = [Point(x-r, pin_length/2 +y), Point(x-r, -pin_length/2 +y)]
    p3c = Point(x-r, y)
    self.set_p3 = p3c
    self.p3 = p3c
    pin = Path(p3, w_ring)
    self.cell.shapes(LayerPinRecN).insert(pin)
    t = Trans(Trans.R0, x-r, y)
    text = Text ("pin3", t)
    shape = self.cell.shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu

    # Pin on the bottom right side:
    p4 = [Point(x+r, y+pin_length/2), Point(x+r,y-pin_length/2)]
    p4c = Point(x+r,y)
    self.set_p4 = p4c
    self.p4 = p4c
    pin = Path(p4, w_ring)
    self.cell.shapes(LayerPinRecN).insert(pin)
    t = Trans(Trans.R0, x+r, y)
    text = Text ("pin4", t)
    shape = self.cell.shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu
    
    # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
    self.cell.shapes(LayerDevRecN).insert(arc_wg_xy(x ,y, r+gap, w*5, 0, 180))



import pya
import math
from pya import *
from SiEPIC.utils import get_technology_by_name

class ebeam_PWB_surface_taper(pya.PCellDeclarationHelper):
  """
  The PCell declaration for the photonic wirebond surface taper
  Author: Becky Lin
  V1: July 30, 2020
  V2: Sept 8, 2020
  V3: Dec 4, 2020
  blin@ece.ubc.ca
  """

  def __init__(self):
    
    # Important: initialize the super class
    super(ebeam_PWB_surface_taper, self).__init__()
    TECHNOLOGY = get_technology_by_name('EBeam')

    self.param("Wtip", self.TypeDouble, "Width of tip (microns)", default = 0.13)
    self.param("Wwaveguide", self.TypeDouble, "width of waveguide (microns)", default = 0.5)
    self.param("edge_offset", self.TypeDouble, "tip to edge gap (microns, recommend min 30um)", default = 30)
    self.param("taper_length", self.TypeDouble, "length of taper (microns, min 45um; max 100um)", default = 65)
    self.param("OxOp_angle", self.TypeInt, "Oxide open interface angle (int)", default = 8)
    self.param("show_deepTrench", self.TypeBoolean, "Show deep trench layer? !Min DT and OxOp is 5um", default = True)
    self.param("ZEP_Process", self.TypeBoolean, "Fabrication using UBC ZEP process", default = False)
    self.param("ZEP_Trench", self.TypeDouble, "Trench width for ZEP process", default = 10)
    self.param("Grouse_Run", self.TypeBoolean, "Fabrication using ANT Grouse Run", default = False)

    # declare the layers
    self.param("silayer", self.TypeLayer, "Si Layer", default=TECHNOLOGY['Si'])
    self.param("pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY['PinRec'])
    self.param("devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY['DevRec'])
    self.param("oxopen", self.TypeLayer, "OxOpen Layer", default=TECHNOLOGY['Oxide open (to BOX)'])
    self.param("textl", self.TypeLayer, "Text Layer", default=TECHNOLOGY['Text'])
    self.param("fibertarget", self.TypeLayer, "Fiber Target Layer", default=TECHNOLOGY['FbrTgt'])
    self.param("ZEP_invert", self.TypeLayer, "31 Si for ZEP inverse", default=LayerInfo(31,0))
    self.param("Grouse_Si_Keepout", self.TypeLayer, "Grouse Si Keep-out Layer", default=LayerInfo(63,0))


  def can_create_from_shape_impl(self):
    return False

  def produce(self, layout, layers, parameters, cell):
    # This is the main part of the implementation: create the layout

    #
    self.cell = cell
    self._param_values = parameters
    self.layout = layout

    # import neccessary libraries
    from math import pi, sin, cos, tan
    from SiEPIC._globals import PIN_LENGTH as pin_length
    import ctypes

    # fetch the parameters
    dbu = self.layout.dbu
    ly = self.layout
    shapes = self.cell.shapes
    
    LayerSiN = ly.layer(self.silayer)
    LayerPinRecN = ly.layer(self.pinrec)
    LayerDevRecN = ly.layer(self.devrec)
    LayerOxOpN = ly.layer(self.oxopen)
    #LayerDTN = ly.layer(self.deeptrench)
    LayerTEXTN = ly.layer(self.textl)
    LayerFbrTgtN = ly.layer(self.fibertarget)
    Layer31SiN = ly.layer(self.ZEP_invert)
    Layer63N = ly.layer(self.Grouse_Si_Keepout)

    w1 = self.Wtip/dbu
    w2 = self.Wwaveguide/dbu
    offset = self.edge_offset
    angle = self.OxOp_angle * (pi/180)
    excl = 750
    notile = 15
    y_oxop = 20
    l = self.taper_length   
    wT = self.ZEP_Trench

    #draw deeptrench layer (to indicate edge)
    pts6 = [Point(0, excl/dbu), Point(-100/dbu,excl/dbu), Point(-100/dbu,-excl/dbu),Point(0,-excl/dbu)]

    #place deeptrench layer
    OxOp_DT_gap = 0

    if self.show_deepTrench:
        #shapes(LayerDTN).insert(Polygon(pts6))
        OxOp_DT_gap = 5
      
        #draw no surface element window for edge coupling
        pts5 = [Point(0, excl/dbu),Point((offset+l+excl)/dbu,excl/dbu),Point((offset+l+excl)/dbu,-excl/dbu),Point(0,-excl/dbu)]

        t = Trans(Trans.R0,0, (excl - 800)/dbu)
        text = Text("edge of chip",t)
        shape = shapes(LayerTEXTN).insert(text)
        shape.text_size = 7/dbu

    else:
        #draw no surface element window for on chip wirebonding
        pts5 = [Point(-(excl+l+150)/dbu, excl/dbu),Point((offset+l+excl)/dbu,excl/dbu),Point((offset+l+excl)/dbu,-excl/dbu),Point(-(excl+l+150)/dbu,-excl/dbu)]

    #draw the taper
    pts = [Point(offset/dbu,w1/2),Point((l+offset)/dbu,w2/2),Point((l+offset)/dbu,-w2/2),Point(offset/dbu,-w1/2)]
    #pts = [Point(offset/dbu,-0.001/dbu),Point((offset-20)/dbu,-0.001/dbu),Point((offset-20)/dbu,0.001/dbu),Point(offset/dbu,0.001/dbu),Point(offset/dbu,w1/2),Point((l+offset)/dbu,w2/2),Point((l+offset)/dbu,-w2/2),Point(offset/dbu,-w1/2)]  

    
    #place the taper
    shapes(LayerSiN).insert(Polygon(pts))

    #draw waveguide section
    Lw2 = 15
    pts7 = [Point((offset+l)/dbu, (w2/2)),Point((offset+l+Lw2)/dbu,(w2/2)),Point((offset+l+Lw2)/dbu,(-w2/2)),Point((offset+l)/dbu,(-w2/2))]
    
    #place waveguide section
    shapes(LayerSiN).insert(Polygon(pts7))

    #draw alignment marker
    pts1 = [Point((0+offset)/dbu,12.4/dbu),Point((0+offset)/dbu,9/dbu),Point((3.5+offset)/dbu,9/dbu),Point((3.5+offset)/dbu,12.5/dbu),Point((0.1+offset)/dbu,12.5/dbu),Point((0+offset)/dbu,12.6/dbu),Point((0+offset)/dbu,16/dbu),Point((-3.5+offset)/dbu,16/dbu),Point((-3.5+offset)/dbu,12.5/dbu),Point((-0.1+offset)/dbu,12.5/dbu)]
    
    pts2 = [Point((0+50+offset)/dbu,12.4/dbu),Point((0+50+offset)/dbu,9/dbu),Point((3.5+50+offset)/dbu,9/dbu),Point((3.5+50+offset)/dbu,12.5/dbu),Point((0.1+50+offset)/dbu,12.5/dbu),Point((0+50+offset)/dbu,12.6/dbu),Point((0+50+offset)/dbu,16/dbu),Point((-3.5+50+offset)/dbu,16/dbu),Point((-3.5+50+offset)/dbu,12.5/dbu),Point((-0.1+50+offset)/dbu,12.5/dbu)]
    
    #place alignment marker
    shapes(LayerSiN).insert(Polygon(pts1))
    shapes(LayerSiN).insert(Polygon(pts2))

    #draw oxide open window
    dy_oxop = 5
    dx = dy_oxop*tan(angle)
    ddx = sin(angle)
    ddy = cos(angle)
    #pts3 = [Point(OxOp_DT_gap/dbu,y_oxop/dbu),Point((offset+l+10+dx-1)/dbu,dy_oxop/dbu),Point((offset+l+10+dx-ddx)/dbu,(y_oxop-ddy)/dbu),Point((offset+l+10-dx)/dbu,-(y_oxop-10)/dbu),Point(OxOp_DT_gap/dbu,-(y_oxop-10)/dbu)]
    
    pts3 = [Point(OxOp_DT_gap/dbu,y_oxop/dbu),Point((offset+l+10+dx-1)/dbu,y_oxop/dbu),Point((offset+l+10+dx-1)/dbu,dy_oxop/dbu),Point((offset+l+10-dx-1)/dbu,-dy_oxop/dbu),Point((offset+l+10-dx-1)/dbu,-(y_oxop-10)/dbu),Point(OxOp_DT_gap/dbu,-(y_oxop-10)/dbu)]

    #place oxide open window
    shapes(LayerOxOpN).insert(Polygon(pts3))

    #draw devrec box
    pts4 = [Point(0,y_oxop/dbu),Point((offset+l+10+dx)/dbu,y_oxop/dbu),Point((offset+l+10+dx)/dbu,-(y_oxop-10)/dbu),Point(0,-(y_oxop-10)/dbu)]
    
    #place devrec box
    shapes(LayerDevRecN).insert(Polygon(pts5))

    #EXCLUSION LAYERS################################################################

    #place no surface element window
    #shapes(LayerTEXTN).insert(Polygon(pts5))
    
    #place no surface element text warning
    t = Trans(Trans.R0,0, (excl-730)/dbu)
    text = Text("Recommend no surface elements \n(e.g. metals pads, grating couplers, oxide open, etc) \nwithin this DevRec box.\nSurface elements in this box may be covered by cladding photoresist",t)
    shape = shapes(LayerTEXTN).insert(text) 
    shape.text_size = 7/dbu

    #ZEP PROCESS#####################################################################

    #draw ZEP process, expanded taper trench used during layer inverting
    #path1 = Path([Point(offset,w2/2),Point((l+offset),w2/2),Point((l+offset),-w2/2),Point(offset,-w2/2)],1)
    #pts8 = [Point(offset/dbu,w2/2),Point((l+offset)/dbu,w2/2),Point((l+offset)/dbu,-w2/2),Point(offset/dbu,-w2/2)]

    #if self.ZEP_Process == True:
      #shapes(Layer31SiN).insert(Polygon(pts8))
      #x = (offset + l)/dbu
      #t = Trans(Trans.R0, x, 0)
      #path1_t = pin.transformed(t)
      #shapes(LayerSiN).insert(path1_t)

    ZEPTrench_pts = [Point((offset+65)/dbu,-wT/dbu),Point((offset-25)/dbu,-wT/dbu),Point((offset-25)/dbu,wT/dbu),Point((offset+65)/dbu,wT/dbu)]
    if self.ZEP_Process == True:
        #place the extended Si Keepout at the tip of taper
        shapes(Layer63N).insert(Polygon(ZEPTrench_pts))

    #GROUSE SI KEEPOUT ##############################################################

    #draw the extended Si Keepout at the tip of taper
    grouseRun_pts = [Point(offset/dbu,-2.175/dbu),Point((offset-25)/dbu,-2.175/dbu),Point((offset-25)/dbu,2.175/dbu),Point(offset/dbu,2.175/dbu)]
    if self.Grouse_Run == True:

        #place the extended Si Keepout at the tip of taper
        shapes(Layer63N).insert(Polygon(grouseRun_pts))

    #################################################################################

    #draw and place pin on the waveguide:
    x = (offset + l + Lw2)/dbu
    t = Trans(Trans.R0, x, 0)
    pin = Path([Point(-pin_length/2,0),Point(pin_length/2,0)], w2)
    pin_t = pin.transformed(t)
    shapes(LayerPinRecN).insert(pin_t)
    text = Text("pin2",t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu

    def circle(x,y,r):
        npts = 180
        theta = 2*math.pi/npts
        pts = []
        for i in range(0,npts):
          pts.append(Point.from_dpoint(pya.DPoint((x+r*math.cos(i*theta))/1,(y+r*math.sin(i*theta))/1)))
        return pts
   
    #draw fibre target circle
    align_circle = circle(offset/dbu,0,2/dbu)

    #place fibre target circle
    shapes(LayerFbrTgtN).insert(Polygon(align_circle))

  def display_text_impl(self):
    #Provide a descriptive text for the cell
    if self.show_deepTrench:
        return "PWB_taper (tip_width=" + ('%.2f' % self.Wtip) + ",waveguide_width=" + ('%.2f' % self.Wwaveguide) + ",taper_length=" + ('%.1f' % self.taper_length) + ",Angle=" + ('%s' % self.OxOp_angle) + ",edge_offset=" + ('%.2f' %self.edge_offset) + ")"
    else:
        return "PWB_taper (tip_width=" + ('%.2f' % self.Wtip) + ",waveguide_width=" + ('%.2f' % self.Wwaveguide) + ",taper_length=" + ('%.1f' % self.taper_length) + ",Angle=" + ('%s' % self.OxOp_angle) + ")"      



from pya import *
import pya

class Waveguide_SBend(pya.PCellDeclarationHelper):
  """
  Input: 
  """
  def __init__(self):


    # Important: initialize the super class
    super(Waveguide_SBend, self).__init__()
    from SiEPIC.utils import get_technology_by_name    
    TECHNOLOGY = get_technology_by_name('EBeam')

    # declare the parameters
    self.param("length", self.TypeDouble, "Waveguide length", default = 10.0)     
    self.param("height", self.TypeDouble, "Waveguide offset height", default = 2)     
    self.param("wg_width", self.TypeDouble, "Waveguide width (microns)", default = 0.5)     
#    self.param("radius", self.TypeDouble, "Waveguide bend radius (microns)", default = 5)     
    self.param("layer", self.TypeLayer, "Layer", default = TECHNOLOGY['Si'])
    self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
    self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])


  def display_text_impl(self):
    # Provide a descriptive text for the cell
    return "Waveguide_SBend_%s-%.3f" % \
    (self.length, self.wg_width)
  
  def coerce_parameters_impl(self):
    pass


  def can_create_from_shape(self, layout, shape, layer):
    return False
    
  def produce_impl(self):
  
    # fetch the parameters
    dbu = self.layout.dbu
    ly = self.layout
    shapes = self.cell.shapes

    from SiEPIC.utils.layout import layout_waveguide_sbend, layout_waveguide_sbend_bezier
    from SiEPIC.extend import to_itype

    LayerSi = self.layer
    LayerSiN = ly.layer(LayerSi)
    LayerPinRecN = ly.layer(self.pinrec)
    LayerDevRecN = ly.layer(self.devrec)

    length = to_itype(self.length,dbu)
    w = to_itype(self.wg_width,dbu)
    h = to_itype(self.height,dbu)
   
#    waveguide_length = layout_waveguide_sbend(self.cell, LayerSiN, pya.Trans(Trans.R0, 0,0), w, r, h, length)
    waveguide_length = to_itype(layout_waveguide_sbend_bezier(self.cell, LayerSiN, Trans(), self.wg_width, self.wg_width, self.height,self.length), dbu)
    
    from SiEPIC._globals import PIN_LENGTH as pin_length

    # Pins on the waveguide:
    x = length
    t = Trans(Trans.R0, x,h)
    pin = Path([Point(-pin_length/2,0), Point(pin_length/2,0)], w)
    pin_t = pin.transformed(t)
    shapes(LayerPinRecN).insert(pin_t)
    text = Text ("pin2", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = to_itype(0.4,dbu)
    shape.text_halign = 2

    x = 0
    t = Trans(Trans.R0, x,0)
    pin = Path([Point(pin_length/2,0), Point(-pin_length/2,0)], w)
    pin_t = pin.transformed(t)
    shapes(LayerPinRecN).insert(pin_t)
    text = Text ("pin1", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = to_itype(0.4,dbu)


    # Compact model information
    t = Trans(Trans.R0, 0, 0)
    text = Text ('Lumerical_INTERCONNECT_library=Design kits/EBeam', t)
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = to_itype(0.1,dbu)
    t = Trans(Trans.R0, 0, w*2)
    text = Text ('Component=ebeam_wg_integral_1550', t)
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = to_itype(0.1,dbu)
    t = Trans(Trans.R0, 0, -w*2)
    text = Text \
      ('Spice_param:wg_length=%.3fu wg_width=%.3fu' %\
      (waveguide_length*dbu, self.wg_width), t )
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = to_itype(0.1,dbu)
#    t = Trans(Trans.R0, 0, -w*3)
#    text = Text ('Extra length = %.4fu, Shortest length = %.4fu' % (straight_l*dbu, (length-2*straight_l)*dbu), t )
#    shape = shapes(LayerDevRecN).insert(text)
#    shape.text_size = 0.1/dbu

    # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
    box1 = Box(0, min(-w*3,h-w*3), length, max(w*3,h+w*3))
    shapes(LayerDevRecN).insert(box1)
    
#    print('2020/11/28 update')



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name

class Grating_Coupler_FSWGC(pya.PCellDeclarationHelper):
  """
  Focusing Sub-wavelength grating coupler PCell implementation.
  Analytical design based on "Grating Coupler Design Based on Silicon-On-Insulator", Yun Wang (2013). Master's Thesis, University of British Columbia, Canada
  Some PCell implementation adapted from the SiEPIC_EBeam library by Dr. Lukas Chrostowski, University of British Columbia, Canada
 
  Separate modelling (e.g. Lumerical MODE) is required to determine the "grating effective index" parameter for a given device layer thickness,
  cladding type, and period/duty cycle/fill factor.

  Script written by Timothy Richards (Simon Fraser University, BC, Canada) and Adam DeAbreu (Simon Fraser University, BC, Canada)

  Changelog

  2017-07-07 - initial publish
  2017-07-07 - change library & component names; commit to github

  TO-DO:
  - implement mode solver here, or call Lumerical MODE to calculate
  - needs to be verified experimentally
  - default parameters are not correct; compare with "ebeam_gc_te1550"

  Input: 
  
  """

  def __init__(self):

    # Important: initialize the super class
    super(Grating_Coupler_FSWGC, self).__init__()
    TECHNOLOGY = get_technology_by_name('EBeam')

    # declare the parameters  
    self.param("wavelength", self.TypeDouble, "Design Wavelength (micron)", default = 1.55)  
    self.param("n_t", self.TypeDouble, "Cladding Index", default = 1.0)
    self.param("n_e", self.TypeDouble, "Grating Effective Index", default = 2.75618)
    self.param("angle_e", self.TypeDouble, "Taper Angle (deg)", default = 35.0)
    self.param("grating_length", self.TypeDouble, "Grating Length (micron)", default = 15.0)
    self.param("taper_length", self.TypeDouble, "Taper Length (micron)", default = 19.0)
    self.param("dc", self.TypeDouble, "Duty Cycle", default = 0.488193)
    self.param("period", self.TypeDouble, "Grating Period", default = 1.18939)
    self.param("ff", self.TypeDouble, "Fill Factor", default = 0.244319)
    self.param("t", self.TypeDouble, "Waveguide Width (micron)", default = 0.5)
    self.param("theta_c", self.TypeDouble, "Insertion Angle (deg)", default = -31.0)
    
    # Layer parameters
    self.param("layer", self.TypeLayer, "Layer", default = TECHNOLOGY['Si'])
    self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
    self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])

  def display_text_impl(self):
    # Provide a descriptive text for the cell
    return "Grating_Coupler_FSWGC_%.1f-%.2f-%.2f-%.2f-%.2f-%.2f-%.2f-%.2f" % \
    (self.wavelength, self.theta_c, self.period, self.dc, self.ff, self.angle_e, self.taper_length, self.t)
    
#    return "temporary placeholder"
    
  def coerce_parameters_impl(self):
    pass

  def can_create_from_shape(self, layout, shape, layer):
    return False
    
  def produce_impl(self):
    # fetch the parameters
    dbu = self.layout.dbu
    ly = self.layout
    shapes = self.cell.shapes

    LayerSi = self.layer
    LayerSiN = ly.layer(LayerSi)
    LayerSiSPN = ly.layer(LayerSi)
    LayerPinRecN = ly.layer(self.pinrec)
    LayerDevRecN = ly.layer(self.devrec)

    from math import pi, cos, sin, log, sqrt, tan
    from SiEPIC.utils import points_per_circle
    
    lambda_0 = self.wavelength                   ##um wavelength of light
#    pin_length =0.0                 ##um extra nub for the waveguide attachment
    
    # Geometry
    wh = self.period*self.dc                   ##thick grating
    wl = self.ff*(self.period - wh)            ## thin grating
    spacing = (self.period - wh - wl)/2   ##space between thick and thin
    
    gc_number = int(round(self.grating_length/self.period)) ##number of periods
    e = self.n_t*sin((pi/180)*self.theta_c)/self.n_e
    N = round(self.taper_length*(1+e)*self.n_e/lambda_0) ##allows room for the taper

    start = (pi - (pi/180)*self.angle_e/2)
    stop = (pi + (pi/180)*self.angle_e/2)
              
    # Draw coupler grating.
    for j in range(gc_number):

        # number of points in the arcs:
        # calculate such that the vertex & edge placement error is < 0.5 nm.
        #   see "SiEPIC_EBeam_functions - points_per_circle" for more details
        radius = N*lambda_0 / ( self.n_e*( 1 - e )) + j*self.period + spacing
        seg_points = int(points_per_circle(radius/dbu)/360.*self.angle_e) # number of points grating arc
        theta_up = []
        for m in range(seg_points+1):    
          theta_up = theta_up + [start + m*(stop-start)/seg_points]
        theta_down = theta_up[::-1]

        ##small one
        r_up = []
        r_down = []
        for k in range(len(theta_up)):
          r_up = r_up + [N*lambda_0 / ( self.n_e*( 1 - e*cos(float(theta_up[k])) )) + j*self.period + spacing]
        r_down = r_up[::-1]
    
        xr = []
        yr = []
        for k in range(len(theta_up)):
          xr = xr + [r_up[k]*cos(theta_up[k])]
          yr = yr + [r_up[k]*sin(theta_up[k])]
    
        xl = []
        yl = []
        for k in range(len(theta_down)):
          xl = xl + [(r_down[k] + wl)*cos(theta_down[k])]
          yl = yl + [(r_down[k] + wl)*sin(theta_down[k])]
    
        x = xr + xl
        y = yr + yl
    
        pts = []
        for i in range(len(x)):
            pts.append(Point.from_dpoint(DPoint(x[i]/dbu, y[i]/dbu)))
        #small_one = core.Boundary(points)
        
        polygon = Polygon(pts)
        if self.ff > 0:
            shapes(LayerSiN).insert(polygon)
        
        ##big one
        r_up = []
        r_down = []
        for k in range(len(theta_up)):
          r_up = r_up + [N*lambda_0 / ( self.n_e*( 1 - e*cos(float(theta_up[k])) )) + j*self.period + 2*spacing+ wl]
        r_down = r_up[::-1]
    
        xr = []
        yr = []
        for k in range(len(theta_up)):
          xr = xr + [r_up[k]*cos(theta_up[k])]
          yr = yr + [r_up[k]*sin(theta_up[k])]
    
        xl = []
        yl = []
        for k in range(len(theta_down)):
          xl = xl + [(r_down[k] + wh)*cos(theta_down[k])]
          yl = yl + [(r_down[k] + wh)*sin(theta_down[k])]
    
        x = xr + xl
        y = yr + yl
        
        pts = []
        for i in range(len(x)):
            pts.append(Point.from_dpoint(DPoint(x[i]/dbu, y[i]/dbu)))
        
        polygon = Polygon(pts)
        shapes(LayerSiN).insert(polygon)
      
    # Taper section
    r_up = []
    r_down = []
    for k in range(len(theta_up)):  
      r_up = r_up + [N*lambda_0 / ( self.n_e*( 1 - e*cos(float(theta_up[k])) ))]
    r_down = r_up[::-1]
     
    xl = []
    yl = []  
    for k in range(len(theta_down)):
      xl = xl + [(r_down[k])*cos(theta_down[k])]
      yl = yl + [(r_down[k])*sin(theta_down[k])]
      
    yr = [self.t/2., self.t/2., -self.t/2., -self.t/2.]
    
    yl_abs = []
    for k in range(len(yl)):
      yl_abs = yl_abs + [abs(yl[k])]
    
    y_max = max(yl_abs)
    iy_max = yl_abs.index(y_max)
      
    L_o = (y_max - self.t/2)/tan((pi/180)*self.angle_e/2)
      
    xr = [L_o+xl[iy_max], 0, 0, L_o+xl[iy_max]]
     
    x = xr + xl
    y = yr + yl  

    pts = []
    for i in range(len(x)):
      pts.append(Point.from_dpoint(DPoint(x[i]/dbu, y[i]/dbu)))
    
    polygon = Polygon(pts)
    shapes(LayerSiN).insert(polygon)


    # Pin on the waveguide:
    from SiEPIC._globals import PIN_LENGTH as pin_length
    x = 0
    t = Trans(Trans.R0, x,0)
    pin = Path([Point(-pin_length/2,0), Point(pin_length/2,0)], self.t/dbu)
    pin_t = pin.transformed(t)
    shapes(LayerPinRecN).insert(pin_t)
    text = Text ("pin1", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu


    # Device recognition layer 
    yr = sin(start) * (N*lambda_0 / ( self.n_e*( 1 - e*cos(float(start)) )) + gc_number*self.period + spacing)
    box1 = Box(-(self.grating_length+self.taper_length)/dbu-pin_length*2, yr/dbu, 0, -yr/dbu)
    shapes(LayerDevRecN).insert(box1)



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name

class ebeam_irph_wg(pya.PCellDeclarationHelper):
  """
  The PCell declaration for the IRPH straight waveguide.
  
  Authors: Jaspreet Jhoja, Lukas Chrostowski
  """

  def __init__(self):

    # Important: initialize the super class
    super(ebeam_irph_wg, self).__init__()
    TECHNOLOGY = get_technology_by_name('EBeam')

    # declare the parameters
    self.param("w", self.TypeDouble, "Waveguide Width", default = 0.5)
    self.param("w_rib", self.TypeDouble, "Waveguide Rib Width", default = 24)
    self.param("length", self.TypeDouble, "Waveguide length", default = 100)
    self.param("n_w", self.TypeDouble, "N Width", default = 22)
    self.param("npp_w", self.TypeDouble, "N++ Width (um)", default = 10)
    self.param("npp_dw", self.TypeDouble, "N++ Edge to Si Edge (um)", default = 2)
    self.param("vc_dw", self.TypeDouble, "VC Edge to Si Edge (um)", default = 5)
    self.param("vc_w", self.TypeDouble, "VC Width", default = 5)
    self.param("m_dw", self.TypeDouble, "Metal edge to Si Edge (um)", default = 4)
    self.param("m_w", self.TypeDouble, "Metal Width", default = 10)
    self.param("overlay", self.TypeDouble, "Overlay accuracy (optical litho) (um)", default = 1)
    self.param("overlay_ebl", self.TypeDouble, "Overlay accuracy (EBL) (um)", default = 0.05)
    self.param("io_wg_type", self.TypeBoolean, "I/O waveguide type, 0 - strip; 1 - rib", default = 0)
        
    self.param("silayer", self.TypeLayer, "Si Layer", default = TECHNOLOGY['Si'])
    self.param("siriblayer", self.TypeLayer, "Si rib Layer", default = TECHNOLOGY['Si - 90 nm rib'])
    self.param("nlayer", self.TypeLayer, "N Layer", default = TECHNOLOGY['Si N'])
    self.param("npplayer", self.TypeLayer, "N++ Layer", default = TECHNOLOGY['Si N++'])
    self.param("vclayer", self.TypeLayer, "VC Layer", default = TECHNOLOGY['VC'])
    self.param("mlayer", self.TypeLayer, "Metal Layer", default = TECHNOLOGY['M2_router'])
    self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
    self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])
    self.param("textl", self.TypeLayer, "Text Layer", default = LayerInfo(10, 0))

    
  def display_text_impl(self):
    # Provide a descriptive text for the cell
    return "ebeam_irph_wg" # (R=" + ('%.3f' % self.r) + ",g=" + ('%g' % (1000*self.g)) + ")"

  def can_create_from_shape_impl(self):
    return False
    
  def produce_impl(self):
    # This is the main part of the implementation: create the layout

    from math import pi, cos, sin
    from SiEPIC.utils import arc_wg, arc_wg_xy
    from SiEPIC._globals import PIN_LENGTH
    from SiEPIC.extend import to_itype

    # fetch the parameters
    dbu = self.layout.dbu
    ly = self.layout
    shapes = self.cell.shapes
    
    LayerSiN = ly.layer(self.silayer)
    LayerSiRibN = ly.layer(self.siriblayer)
    LayerNN = ly.layer(self.nlayer)
    LayerNPPN = ly.layer(self.npplayer)
    LayerVCN = ly.layer(self.vclayer)
    LayerMN = ly.layer(self.mlayer)
    LayerPinRecN = ly.layer(self.pinrec)
    LayerDevRecN = ly.layer(self.devrec)
    TextLayerN = ly.layer(self.textl)

    w = to_itype(self.w,dbu)
    w_rib = to_itype(self.w_rib,dbu)
    length = to_itype(self.length,dbu)
    n_w = to_itype(self.n_w,dbu)
    npp_w=to_itype(self.npp_w,dbu)
    npp_dw=to_itype(self.npp_dw,dbu)
    vc_w = to_itype(self.vc_w,dbu)
    vc_dw = to_itype(self.vc_dw,dbu)
    m_dw = to_itype(self.m_dw,dbu)
    m_w = to_itype(self.m_w,dbu)
    overlay = to_itype(self.overlay, dbu)
    overlay_ebl = to_itype(self.overlay_ebl, dbu)
    
    
    if self.io_wg_type == 0:
      ## define strip to slab transfer taper parameters
      N = 100; # Number of points for the input/output slab taper
      order = 3; # input/output slab taper curve 
      in_taper = to_itype(5,dbu); # length of taper
      taper_w = to_itype(2,dbu)
      # add input slab taper
      pts = [];
      for i in range(0,N + 1):
        pts.append(Point(- in_taper + in_taper/N * i, (w-overlay_ebl*2)/2 + ((taper_w - (w-overlay_ebl*2))/(N**order))*(i**order)))
      for i in range(0,N + 1):
        pts.append(Point(-in_taper + in_taper/N * (N - i), - (w-overlay_ebl*2)/2 - ((taper_w - (w-overlay_ebl*2))/(N**order))*((N-i)**order)))
      self.cell.shapes(LayerSiRibN).insert(Polygon(pts))
      # add output slab taper
      pts = [];
      for i in range(0,N + 1):
        pts.append(Point(length+ in_taper - in_taper/N * i, (w-overlay_ebl*2)/2 + ((taper_w - (w-overlay_ebl*2))/(N**order))*(i**order)))
      for i in range(0,N + 1):
        pts.append(Point(length+in_taper - in_taper/N * (N - i), - (w-overlay_ebl*2)/2 - ((taper_w - (w-overlay_ebl*2))/(N**order))*((N-i)**order)))
      self.cell.shapes(LayerSiRibN).insert(Polygon(pts))
    else:
      in_taper = 0; # length of taper
      taper_w = 0; # width of taper

    #draw the waveguide
    xtop = 0 - in_taper
    ytop = -1*(w/2)
    xbottom = length + in_taper
    ybottom = 1*(w/2)
    wg1 = Box(xtop, -w/2, xbottom, w/2)
    shapes(LayerSiN).insert(wg1)


    # Pins on the bottom waveguide side:    
    pin = Path([Point(xtop+PIN_LENGTH/2, 0), Point(xtop-PIN_LENGTH/2, 0)], w)
    shapes(LayerPinRecN).insert(pin)
    text = Text ("pin1", Trans(Trans.R0, xtop, 0))
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu      

    pin = Path([Point(xbottom-PIN_LENGTH/2, 0), Point(xbottom+PIN_LENGTH/2, 0)], w)
    shapes(LayerPinRecN).insert(pin)
    text = Text ("pin2", Trans(Trans.R0, xbottom, 0))
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu      
  
    #draw N
    b = Box(0, -n_w/2, length, n_w/2)
    shapes(LayerNN).insert(b)
    
    #draw NPP
    b = Box(0, npp_dw+w/2, length, npp_dw+w/2+npp_w)
    shapes(LayerNPPN).insert(b)
    b = Box(0, -npp_dw-w/2, length, -npp_dw-w/2-npp_w)
    shapes(LayerNPPN).insert(b)

    #draw via
    b = Box(overlay, vc_dw+w/2, length-overlay, vc_dw+w/2+vc_w)
    shapes(LayerVCN).insert(b)
    b = Box(overlay, -vc_dw-w/2, length-overlay, -vc_dw-w/2-vc_w)
    shapes(LayerVCN).insert(b)
    
    #draw metal
    b = Box(0, m_dw+w/2, length, m_dw+w/2+m_w)
    shapes(LayerMN).insert(b)
    b = Box(0, -m_dw-w/2, length, -m_dw-w/2-m_w)
    shapes(LayerMN).insert(b)


    #devrec layer
    w_devrec = 2 * max(m_dw+m_w+w/2, vc_dw+w/2+vc_w, npp_dw+w/2+npp_w, w_rib/2)
    pts = [];
    pts = [ Point(0, -w_devrec/2-w), 
            Point(0, -taper_w), 
            Point(0-in_taper, -taper_w), 
            Point(0-in_taper, +taper_w), 
            Point(0, +taper_w), 
            Point(0, w_devrec/2+w), 
            Point(length, w_devrec/2+w), 
            Point(length, +taper_w), 
            Point(length+in_taper, +taper_w), 
            Point(length+in_taper, -taper_w), 
            Point(length, -taper_w), 
            Point(length, -w_devrec/2-w) ]            
    self.cell.shapes(LayerDevRecN).insert(Polygon(pts))


    # Si rib layer
    b = Box(0, w_rib/2, length, -w_rib/2)
    shapes(LayerSiRibN).insert(b)
    
    t = Trans(Trans.R0, 0,0)
    text = Text ('Component=ebeam_irph_wg', t)
    shape = shapes(LayerDevRecN).insert(text)
    # shape.text_size = self.r*0.07/dbu
    
#    print("Done drawing the layout for - ebeam_irph_wg: %.3f-%g" % ( self.r, self.g))
  


import pya
from pya import *
from SiEPIC.utils import get_technology_by_name

class Contra_DC_custom(pya.PCellDeclarationHelper):
  """
  Author:   Mustafa Hammood 
            Mustafa@siepic.com
  """

  def __init__(self):

    # Important: initialize the super class
    super(Contra_DC_custom, self).__init__()
    TECHNOLOGY = get_technology_by_name('EBeam')

    # declare the parameters
    self.param("number_of_periods1", self.TypeInt, "Number of grating periods", default = 300)
    self.param("grating_period1", self.TypeDouble, "Grating period (inner)", default = 0.318)
    self.param("grating_period2", self.TypeDouble, "Grating period (outer)", default = 0.318)
    self.param("gap", self.TypeDouble, "Gap (microns)", default = 0.15)
    self.param("corrugation_width1outer", self.TypeDouble, "Waveguide 1 Corrugration width (outer)", default = 0.03)
    self.param("corrugation_width1inner", self.TypeDouble, "Waveguide 1 Corrugration width (inner)", default = 0.04)
    self.param("corrugation_width2outer", self.TypeDouble, "Waveguide 2 Corrugration width (inner)", default = 0.01)
    self.param("corrugation_width2inner", self.TypeDouble, "Waveguide 2 Corrugration width (outer)", default = 0.01)
    self.param("wg1_width", self.TypeDouble, "Waveguide 1 width", default = 0.45)
    self.param("wg2_width", self.TypeDouble, "Waveguide 2 width", default = 0.55)          
    self.param("index", self.TypeDouble, "Gaussian Index", default = 2.8)
    self.param("layer", self.TypeLayer, "Layer", default = TECHNOLOGY['Si'])
    self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
    self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])
    self.param("textl", self.TypeLayer, "Text Layer", default = LayerInfo(10, 0))

  def display_text_impl(self):
    # Provide a descriptive text for the cell
    return "Contra_DC_dualBragg%s-%.3f" % \
    (self.number_of_periods1, self.grating_period1)
  
  def coerce_parameters_impl(self):
    pass

  def can_create_from_shape(self, layout, shape, layer):
    return False
    
  def produce_impl(self):
    import math
    
    # fetch the parameters
    dbu = self.layout.dbu
    ly = self.layout
    shapes = self.cell.shapes

    LayerSi = self.layer
    LayerSiN = ly.layer(LayerSi)
    LayerPinRecN = ly.layer(self.pinrec)
    LayerDevRecN = ly.layer(self.devrec)

    from SiEPIC.extend import to_itype
    
    # Draw the Bragg grating (bottom):
    width_period1 = to_itype(self.grating_period1/2,dbu)
    width_period2 = to_itype(self.grating_period2/2,dbu)
    
    grating_period1 = to_itype(self.grating_period1,dbu)
    grating_period2 = to_itype(self.grating_period2,dbu)

    w1 = to_itype(self.wg1_width,dbu)
    w2 = to_itype(self.wg2_width,dbu)
    
    GaussianIndex = self.index

    N1 = self.number_of_periods1
    
    vertical_offset1 = to_itype(self.gap/2 + self.wg1_width/2, dbu)
    vertical_offset2 = to_itype(-self.gap/2 - self.wg2_width/2, dbu)
    
    for i in range(0,self.number_of_periods1):
      x1 = to_itype(i * self.grating_period1,dbu)
      x2 = to_itype(i * self.grating_period2,dbu)
      
      profileFunction = math.exp( -0.5*(2*GaussianIndex*(i-N1/2)/(N1))**2 )
      
      profile_outer = to_itype(self.corrugation_width1outer,dbu)*profileFunction
      profile_inner = to_itype(self.corrugation_width1inner,dbu)*profileFunction
      
      box1outer = Box(x1, vertical_offset1+w1/2-profile_outer/2, x1 + width_period1, vertical_offset1)
      box2outer = Box(x1+width_period1, vertical_offset1+w1/2+profile_outer/2, x1 + 2*width_period1, vertical_offset1)
      
      box1inner = Box(x2, vertical_offset1-w1/2-profile_inner/2, x2 + width_period2, vertical_offset1)
      box2inner = Box(x2+width_period2, vertical_offset1-w1/2+profile_inner/2, x2 + 2*width_period2, vertical_offset1)
      
      shapes(LayerSiN).insert(box1outer)
      shapes(LayerSiN).insert(box2outer)

      shapes(LayerSiN).insert(box1inner)
      shapes(LayerSiN).insert(box2inner)
      
      length1 = x1 + grating_period1
      length2 = x2 + grating_period2
    
    if grating_period1 > grating_period2:
      box_inner = Box(length2, vertical_offset1, length1, vertical_offset1-w1/2)
      shapes(LayerSiN).insert(box_inner)
      length = x1 + grating_period1
    else:
      box_outer = Box(length1, vertical_offset1, length2, vertical_offset1+w1/2)
      shapes(LayerSiN).insert(box_outer)
      length = x2 + grating_period2

    # lower waveguide
    for i in range(0,self.number_of_periods1):
      x1 = to_itype(i * self.grating_period1,dbu)
      x2 = to_itype(i * self.grating_period2,dbu)
      
      profileFunction = math.exp( -0.5*(2*GaussianIndex*(i-N1/2)/(N1))**2 )
      
      profile_outer = to_itype(self.corrugation_width2outer,dbu)*profileFunction
      profile_inner = to_itype(self.corrugation_width2inner,dbu)*profileFunction
      
      box1outer = Box(x2, vertical_offset2+w2/2+profile_outer/2, x2 + width_period2, vertical_offset2)
      box2outer = Box(x2+width_period2, vertical_offset2+w2/2-profile_outer/2, x2 + 2*width_period2, vertical_offset2)
      
      box1inner = Box(x1, vertical_offset2-w2/2+profile_inner/2, x1 + width_period1, vertical_offset2)
      box2inner = Box(x1+width_period1, vertical_offset2-w2/2-profile_inner/2, x1 + 2*width_period1, vertical_offset2)
      
      shapes(LayerSiN).insert(box1outer)
      shapes(LayerSiN).insert(box2outer)

      shapes(LayerSiN).insert(box1inner)
      shapes(LayerSiN).insert(box2inner)
      
      length1 = x1 + grating_period1
      length2 = x2 + grating_period2
    
    if grating_period1 > grating_period2:
      box_inner = Box(length2, vertical_offset2, length1, vertical_offset2+w2/2)
      shapes(LayerSiN).insert(box_inner)
      length = x1 + grating_period1
    else:
      box_outer = Box(length1, vertical_offset2, length2, vertical_offset2-w2/2)
      shapes(LayerSiN).insert(box_outer)
      length = x2 + grating_period2

    # Create the pins on the waveguides, as short paths:
    from SiEPIC._globals import PIN_LENGTH as pin_length

    w = to_itype(self.wg2_width,dbu)
    t = Trans(Trans.R0, 0,vertical_offset2)
    pin = Path([Point(pin_length/2, 0), Point(-pin_length/2, 0)], w)
    pin_t = pin.transformed(t)
    shapes(LayerPinRecN).insert(pin_t)
    text = Text ("pin1", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu
    
    w = to_itype(self.wg1_width,dbu)
    t = Trans(Trans.R0, 0,vertical_offset1)
    pin = Path([Point(pin_length/2, 0), Point(-pin_length/2, 0)], w)
    pin_t = pin.transformed(t)
    shapes(LayerPinRecN).insert(pin_t)
    text = Text ("pin3", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu

    w = to_itype(self.wg1_width,dbu)
    t = Trans(Trans.R0, length,vertical_offset1)
    pin = Path([Point(-pin_length/2, 0), Point(pin_length/2, 0)], w)
    pin_t = pin.transformed(t)
    shapes(LayerPinRecN).insert(pin_t)
    text = Text ("pin2", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu
    
    w = to_itype(self.wg2_width,dbu)
    t = Trans(Trans.R0, length,vertical_offset2)
    pin = Path([Point(-pin_length/2, 0), Point(pin_length/2, 0)], w)
    pin_t = pin.transformed(t)
    shapes(LayerPinRecN).insert(pin_t)
    text = Text ("pin4", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu

    # Compact model information
    t = Trans(Trans.R0, 0, 0)
    text = Text ('Lumerical_INTERCONNECT_library=Design kits/ebeam', t)
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = 0.1/dbu
    t = Trans(Trans.R0, length/10, 0)
    text = Text ('Component=ebeam_contra_dc', t)
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = 0.1/dbu
    t = Trans(Trans.R0, length/9, 0)
    text = Text \
      ('Spice_param:number_of_periods=%s' %\
      (self.number_of_periods1), t )
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = 0.1/dbu

    # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
    t = Trans(Trans.R0, 0,0)
    path = Path([Point(0, 0), Point(length, 0)], 3*w)
    shapes(LayerDevRecN).insert(path.simple_polygon())



import pya
from pya import *
from SiEPIC.utils import get_technology, get_technology_by_name
from SiEPIC.utils import arc, arc_wg, arc_to_waveguide, points_per_circle#,layout
import math


class swg_fc_test(pya.PCellDeclarationHelper):
  """
  Sub-wavelength-grating fibre coupler PCell litho test structure.

  2017/07/12: Lukas Chrostowski, initial version, based on swg_fc by Tim

  Input: 
  
  """

  def __init__(self):

    # Important: initialize the super class
    super(swg_fc_test, self).__init__()

    # declare the parameters  
    self.param("wavelength", self.TypeDouble, "Design Wavelength (micron)", default = 2.9)  
    self.param("n_t", self.TypeDouble, "Fiber Mode", default = 1.0)
    self.param("n_e", self.TypeDouble, "Grating Index Parameter", default = 3.1)
    self.param("angle_e", self.TypeDouble, "Taper Angle (deg)", default = 20.0)
    self.param("grating_length", self.TypeDouble, "Grating Length (micron)", default = 3.0)
    self.param("taper_length", self.TypeDouble, "Taper Length (micron)", default = 32.0)
    self.param("dc", self.TypeDouble, "Duty Cycle", default = 0.488193)
    self.param("period", self.TypeDouble, "Grating Period", default = 1.18939)
    self.param("ff", self.TypeDouble, "Fill Factor", default = 0.244319)
    self.param("t", self.TypeDouble, "Waveguide Width (micron)", default = 1.0)
    self.param("theta_c", self.TypeDouble, "Insertion Angle (deg)", default = 8.0)
    self.param("fab_error", self.TypeDouble, "Fab Process error max (micron)", default = 0.05)

    
    # Layer parameters
    TECHNOLOGY = get_technology_by_name('EBeam')
    self.param("layer", self.TypeLayer, "Layer", default = TECHNOLOGY['Si'])
    self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
    self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])
    self.param("textl", self.TypeLayer, "Text Layer", default = TECHNOLOGY['Text'])

  def display_text_impl(self):
    # Provide a descriptive text for the cell
    return "swg_fc_test_%.1f-%.2f-%.2f-%.2f-%.2f-%.2f-%.2f-%.2f" % \
    (self.wavelength, self.theta_c, self.period, self.dc, self.ff, self.angle_e, self.taper_length, self.t)
    
#    return "temporary placeholder"
    
  def coerce_parameters_impl(self):
    pass

  def can_create_from_shape(self, layout, shape, layer):
    return False
    
  def produce_impl(self):
    # fetch the parameters
    dbu = self.layout.dbu
    ly = self.layout
    shapes = self.cell.shapes

    LayerSi = self.layer
    LayerSiN = ly.layer(self.layer)
    LayerPinRecN = ly.layer(self.pinrec)
    LayerDevRecN = ly.layer(self.devrec)
    LayerTextN = ly.layer(self.textl)

    from math import pi, cos, sin, log, sqrt, tan
    
    lambda_0 = self.wavelength                   ##um wavelength of light
    pin_length = 0.5                 ##um extra nub for the waveguid attachment
    
    # Geometry
    wh = self.period*self.dc                   ##thick grating
    wl = self.ff*(self.period - wh)            ## thin grating
    spacing = (self.period - wh - wl)/2   ##space between thick and thin
    
    gc_number = int(round(self.grating_length/self.period)) ##number of periods
    gc_number = 3
    e = self.n_t*sin((pi/180)*self.theta_c)/self.n_e
    N = round(self.taper_length*(1+e)*self.n_e/lambda_0) ##allows room for the taper

    start = (pi - (pi/180)*self.angle_e/2)
    stop = (pi + (pi/180)*self.angle_e/2)
              
    # Draw coupler grating.
    for j in range(gc_number):

        # number of points in the arcs:
        # calculate such that the vertex & edge placement error is < 0.5 nm.
        #   see "SiEPIC_EBeam_functions - points_per_circle" for more details
        radius = N*lambda_0 / ( self.n_e*( 1 - e )) + j*self.period + spacing
        seg_points = int(points_per_circle(radius/dbu)/360.*self.angle_e) # number of points grating arc
        theta_up = []
        for m in range(seg_points+1):    
          theta_up = theta_up + [start + m*(stop-start)/seg_points]
        theta_down = theta_up[::-1]
        
        ##small one
        r_up = []
        r_down = []
        rng = range(len(theta_up))
        
        # find the divider to get desired fab error:
        th = min(theta_up)
        div = (2*sin(th)/self.fab_error)*(N*lambda_0 / ( self.n_e*( 1 - e*cos(th) )) + j*self.period + spacing)
        err = (2*sin(th)/div)*(N*lambda_0 / ( self.n_e*( 1 - e*cos(th) )) + j*self.period + spacing)
#        print("div %s, err (double check) %s" % (div, err))

        for k in rng:
          th = theta_up[k]
#          print("%s, %s, %s" % (th, sin(th), 1+sin(th)/10.) )
          r_up = r_up +  [(1-sin(th)/div) *(N*lambda_0 / ( self.n_e*( 1 - e*cos(th) )) + j*self.period + spacing)]
        for k in rng[::-1]:
          th = theta_up[k]
#          print("%s, %s, %s" % (th, sin(th), 1+sin(th)/10.) )
          r_down = r_down +  [(1+sin(th)/div) *(N*lambda_0 / ( self.n_e*( 1 - e*cos(th) )) + j*self.period + spacing)]
    
        xr = []
        yr = []
        for k in range(len(theta_up)):
          xr = xr + [r_up[k]*cos(theta_up[k])]
          yr = yr + [r_up[k]*sin(theta_up[k])]
    
        xl = []
        yl = []
        for k in range(len(theta_down)):
          xl = xl + [(r_down[k] + wl)*cos(theta_down[k])]
          yl = yl + [(r_down[k] + wl)*sin(theta_down[k])]
    
        x = xr + xl
        y = yr + yl
    
        pts = []
        for i in range(len(x)):
            pts.append(Point.from_dpoint(pya.DPoint(x[i]/dbu, y[i]/dbu)))
        #small_one = core.Boundary(points)
        
        polygon = Polygon(pts)
        shapes(LayerSiN).insert(polygon)

        if j==1:         
          # text label dimensions, for minor grating:
          # top 
          shapes(LayerTextN).insert(Text("%0.0f"%((wl+self.fab_error)*1000), Trans(Trans.R0, xl[0]/dbu,yl[0]/dbu))).text_size = 0.2/dbu
          # btm 
          shapes(LayerTextN).insert(Text("%0.0f"%((wl-self.fab_error)*1000), Trans(Trans.R0, xl[-1]/dbu,yl[-1]/dbu))).text_size = 0.2/dbu
          # mid 
          shapes(LayerTextN).insert(Text("%0.0f"%((wl)*1000), Trans(Trans.R0, xl[int(len(theta_up)/2)]/dbu,yl[int(len(theta_up)/2)]/dbu))).text_size = 0.2/dbu
        
        
        ##big one
        r_up = []
        r_down = []
        
        # find the divider to get desired fab error:
        th = min(theta_up)
        div = (2*sin(th)/self.fab_error)*(N*lambda_0 / ( self.n_e*( 1 - e*cos(th) )) + j*self.period + 2*spacing+wl)
        err = (2*sin(th)/div)*(N*lambda_0 / ( self.n_e*( 1 - e*cos(th) )) + j*self.period + 2*spacing+wl)
#        print("div %s, err (double check) %s" % (div, err))

        rng = range(len(theta_up))
        for k in rng:
          th = theta_up[k]
          r_up = r_up +  [(1-sin(th)/div) *(N*lambda_0 / ( self.n_e*( 1 - e*cos(th) )) + j*self.period + 2*spacing+wl)]
        for k in rng[::-1]:
          th = theta_up[k]
          r_down = r_down +  [(1+sin(th)/div) *(N*lambda_0 / ( self.n_e*( 1 - e*cos(th) )) + j*self.period + 2*spacing+wl)]
    
        xr = []
        yr = []
        for k in range(len(theta_up)):
          xr = xr + [r_up[k]*cos(theta_up[k])]
          yr = yr + [r_up[k]*sin(theta_up[k])]
    
        xl = []
        yl = []
        for k in range(len(theta_down)):
          xl = xl + [(r_down[k] + wh)*cos(theta_down[k])]
          yl = yl + [(r_down[k] + wh)*sin(theta_down[k])]
    
        x = xr + xl
        y = yr + yl
        
        pts = []
        for i in range(len(x)):
            pts.append(Point.from_dpoint(pya.DPoint(x[i]/dbu, y[i]/dbu)))
        
        polygon = Polygon(pts)
        shapes(LayerSiN).insert(polygon)
      
        if j==1:         
          # text label dimensions, for major grating:
          # top 
          shapes(LayerTextN).insert(Text("%0.0f"%((wh+self.fab_error)*1000), Trans(Trans.R0, xl[0]/dbu,yl[0]/dbu))).text_size = 0.2/dbu
          # btm 
          shapes(LayerTextN).insert(Text("%0.0f"%((wh-self.fab_error)*1000), Trans(Trans.R0, xl[-1]/dbu,yl[-1]/dbu))).text_size = 0.2/dbu
          # mid 
          shapes(LayerTextN).insert(Text("%0.0f"%((wh)*1000), Trans(Trans.R0, xl[int(len(theta_up)/2)]/dbu,yl[int(len(theta_up)/2)]/dbu))).text_size = 0.2/dbu



import pya
from pya import *


class ebeam_dream_splitter_1x2_te1550_BB(pya.PCellDeclarationHelper):
    """
    The PCell declaration for the ebeam_dream_splitter_1x2_te1550_BB
    
    Authors: Dream Photonics
    """

    def __init__(self):

        # Important: initialize the super class
        super(ebeam_dream_splitter_1x2_te1550_BB, self).__init__()
        from SiEPIC.utils import get_technology_by_name, load_Waveguides_by_Tech
        self.technology_name = 'EBeam'
        TECHNOLOGY = get_technology_by_name(self.technology_name)

        from SiEPIC._globals import KLAYOUT_VERSION
        if KLAYOUT_VERSION >= 28:
            # Button to launch separate window
            self.param("documentation", self.TypeCallback, "Open documentation in web browser")

    def callback(self, layout, name, states):
        ''' Callback for PCell, to launch documentation viewer
                https://www.klayout.de/doc/code/class_PCellDeclaration.html#method9
        '''
        if name == 'documentation':
            url = 'https://github.com/SiEPIC/SiEPIC_EBeam_PDK/wiki/Component-Library-description#dream-photonics-1x2-splitter'
            import webbrowser
            webbrowser.open_new(url)
        
    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "ebeam_dream_splitter_1x2_te1550_BB" 

    def can_create_from_shape_impl(self):
        return False
        
    def produce_impl(self):
        # This is the main part of the implementation: create the layout

        # Load the geometries into KLayout
        import os
        dir_path = os.path.normpath(os.path.join(os.path.dirname(os.path.realpath(__file__)), "../../gds/EBeam_Dream/"))
        filename = os.path.join(dir_path, 'ebeam_dream_splitter_1x2_te1550_BB.gds')
        tech_name = 'EBeam'
        ly2=pya.Layout()
        ly2.read(filename)
        top_cell = ly2.top_cell()
        top_cell.flatten(True)
        self.cell.copy_shapes(top_cell)



from pya import *
import pya

class Waveguide_Bend(pya.PCellDeclarationHelper):
  """
  The PCell declaration for the waveguide bend.
  """

  def __init__(self):

    # Important: initialize the super class
    super(Waveguide_Bend, self).__init__()
    from SiEPIC.utils import get_technology_by_name    
    TECHNOLOGY = get_technology_by_name('EBeam')

    # declare the parameters
    self.param("silayer", self.TypeLayer, "Si Layer", default = TECHNOLOGY['Si'])
    self.param("radius", self.TypeDouble, "Radius", default = 10)
    self.param("wg_width", self.TypeDouble, "Waveguide Width", default = 0.5)
    self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
    self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])
    # hidden parameters, can be used to query this component:
    self.param("p1", self.TypeShape, "DPoint location of pin1", default = Point(-10000, 0), hidden = True, readonly = True)
    self.param("p2", self.TypeShape, "DPoint location of pin2", default = Point(0, 10000), hidden = True, readonly = True)
    

  def display_text_impl(self):
    # Provide a descriptive text for the cell
    return "Waveguide_Bend(R=" + ('%.3f' % self.radius) + ")"

  def can_create_from_shape_impl(self):
    return False


  def produce(self, layout, layers, parameters, cell):
    """
    coerce parameters (make consistent)
    """
    self._layers = layers
    self.cell = cell
    self._param_values = parameters
    self.layout = layout


    # cell: layout cell to place the layout
    # LayerSiN: which layer to use
    # r: radius
    # w: waveguide width
    # length units in dbu

    from math import pi, cos, sin
    from SiEPIC.utils import arc, arc_to_waveguide, points_per_circle, arc_wg

    # fetch the parameters
    dbu = self.layout.dbu
    ly = self.layout
    shapes = self.cell.shapes
    
    LayerSi = self.silayer
    LayerSiN = self.silayer_layer
    LayerPinRecN = ly.layer(self.pinrec)
    LayerDevRecN = ly.layer(self.devrec)
    
    w = int(round(self.wg_width/dbu))
    r = int(round(self.radius/dbu))

    # draw the quarter-circle
    x = -r
    y = r
   # layout_arc_wg_dbu(self.cell, LayerSiN, x, y, r, w, 270, 360)
    t = Trans(Trans.R0,x, y)
    from SiEPIC import __version__ 
    if __version__ > '0.5.0':
      self.cell.shapes(LayerSiN).insert(arc_to_waveguide(arc(r, 270, 360, dbu=dbu), w).transformed(t))
    else:
      self.cell.shapes(LayerSiN).insert(arc_to_waveguide(arc(r, 270, 360), w).transformed(t))
    
    # Create the pins on the waveguides, as short paths:
    from SiEPIC._globals import PIN_LENGTH as pin_length
    
    # Pin on the top side:
    p2 = [Point(0, y-pin_length/2), Point(0, y+pin_length/2)]
    p2c = Point(0, y)
    self.set_p2 = p2c
    self.p2 = p2c
    pin = Path(p2, w)
    shapes(LayerPinRecN).insert(pin)
    t = Trans(Trans.R0, 0, y)
    text = Text ("pin2", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu

    # Pin on the left side:
    p1 = [Point(pin_length/2+x,0), Point(-pin_length/2+x,0)]
    p1c = Point(x,0)
    self.set_p1 = p1c
    self.p1 = p1c
    pin = Path(p1, w)
    shapes(LayerPinRecN).insert(pin)
    t = Trans(Trans.R0, x, 0)
    text = Text ("pin1", t)
    shape = shapes(LayerPinRecN).insert(text)
    shape.text_size = 0.4/dbu

    # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
    t = Trans(Trans.R0,x, y)
    if __version__ > '0.5.0':
      self.cell.shapes(LayerDevRecN).insert(arc_to_waveguide(arc(r, 270, 360,dbu=dbu), w*3).transformed(t))
    else:
      self.cell.shapes(LayerDevRecN).insert(arc_to_waveguide(arc(r, 270, 360), w*3).transformed(t))
    #layout_arc_wg_dbu(self.cell, LayerDevRecN, x, y, r, w*3, 270, 360)

    # Compact model information
    t = Trans(Trans.R0, x+r/10, 0)
    text = Text ("Lumerical_INTERCONNECT_library=Design kits/EBeam", t)
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = r/100
    t = Trans(Trans.R0, x+r/10, r/4)
    text = Text ('Component=ebeam_bend_1550', t)
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = r/100
    t = Trans(Trans.R0, x+r/10, r/2)
    text = Text ('Spice_param:radius=%.3fu wg_width=%.3fu'% (self.radius,self.wg_width), t)
    shape = shapes(LayerDevRecN).insert(text)
    shape.text_size = r/100



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name

class SWG_MultiBox_Ring(pya.PCellDeclarationHelper):

    def __init__(self):

      # Important: initialize the super class
      super(SWG_MultiBox_Ring, self).__init__()
      TECHNOLOGY = get_technology_by_name('EBeam')

      # declare the parameters
      self.param("pitch", self.TypeDouble, "Period [um]", default = 0.24)
      self.param("w", self.TypeDouble, "Waveguide Width [um]", default = 0.18)
      self.param("r", self.TypeDouble, "Radius [um]", default = 35) # minimum radius is 5 um in TE
      self.param("ff", self.TypeDouble, "Duty Cycle", default = 0.75)
      self.param("angle", self.TypeDouble, "Angle", default = 360)
      self.param("gap", self.TypeDouble, "Row Gap [um]", default = 0.06)
      self.param("gap2", self.TypeDouble, "Bus Gap [um]", default = 0.06)
      self.param("row", self.TypeDouble, "Number of Rows", default = 5)
      self.param("busL", self.TypeDouble, "Bus Length [um]", default = 10)
      self.param("taperL", self.TypeDouble, "Taper Length [um]", default = 25)
      #self.param("doublebus", self.TypeBoolean, "Bus Type (Single=False, Double=True)", default = False)
      self.param("layer", self.TypeLayer, "Layer", default = TECHNOLOGY['Si'])
      self.param("pinrec", self.TypeLayer, "PinRec Layer", default = TECHNOLOGY['PinRec'])
      self.param("devrec", self.TypeLayer, "DevRec Layer", default = TECHNOLOGY['DevRec'])
  #    self.param("textl", self.TypeLayer, "Text Layer", default = LayerInfo(10, 0))
      self.param("oxideopen", self.TypeLayer, "Oxide Open Layer", default = TECHNOLOGY['Oxide open (to BOX)'])

    def display_text_impl(self):
      # Provide a descriptive text for the cell
      return "SWG_multibox_%.3f" % \
      (self.pitch)

    def coerce_parameters_impl(self):
      pass

    def can_create_from_shape(self, layout, shape, layer):
      return False

    def produce_impl(self):
      debug = False
      from SiEPIC._globals import PIN_LENGTH
      from SiEPIC.extend import to_itype
      import math
      from pya import DPolygon
      
      pi  = math.pi
      # This is the main part of the implementation: create the layout

      # fetch the parameters
      dbu = self.layout.dbu
      ly = self.layout
      shapes = self.cell.shapes
      LayerSi = self.layer
      LayerSiN = ly.layer(LayerSi)
      LayerPinRecN = ly.layer(self.pinrec)
      LayerDevRecN = ly.layer(self.devrec)

      #variables
      pitch = self.pitch
      w = self.w
      r = self.r
      ff = self.ff
      angle = self.angle
      gap = self.gap
      gap2 = self.gap2
      row = self.row
      busL = self.busL
      taperL = self.taperL
      #doublebus = self.doublebus
      #print(doublebus)
      
      if (r-w/2 <= 0):
        r = 5
        if debug:
            print ('invalid radius, set r to default: 5')
      if row <= 1:
        row = 1
        if debug:
            print ('invalid number of rows, set row to default: 1')
      if busL <= 10:
        busL = taperL*2+10
        if debug:
            print('invalid length of SWG bus waveguide, set length at least 2 times larger than taper length and pluse 10 um')
      #Calculate number of segments
      s1 = pitch*ff #silicon
      s2 = pitch-s1 #gap

      # draw oxide open 
      shapes(ly.layer(self.oxideopen)).insert(
        pya.Box(
            DPoint(-to_itype(r+5,dbu),-to_itype(r+5,dbu)),
            DPoint(to_itype(r+5,dbu),to_itype(r+5,dbu)) ) )
      
      # Draw the Multi-box Ring
      for i in range(0,int(row)): # draw different radius SWG rings
        #calculate best radius
        #const = math.ceil(2*pi*r/(s1+s2))
        const = math.floor(2*pi*r/(s1+s2))
        #if doesn't divide evenly, replace r with best possible r
        if ((2*pi*r)%(s1+s2) != 0):
          r = const*(s1+s2)/(2*pi)
          if debug:
              print('r adjusted to '+str(r)+'um to fit periods perfectly.')
        
        
        theta1 = math.atan(s1/r)
        theta2 = math.atan(s2/r)
        nSeg = int(math.floor(angle/(math.degrees(theta1)+math.degrees(theta2)))) #how many segments to have
        si_first = True #for alternating between silicon and gap
        j=0 #index of how many silicon thetas 
        jj = 0 #index of how many gap thetas
        ORDER = True #ordering of the coordinates for polygon drawing
       
        xo = [(r-w/2)*math.cos(0)]
        yo = [(r-w/2)*math.sin(0)]
        xo.append((r+w/2)*math.cos(0))
        yo.append((r+w/2)*math.sin(0))
        
              
        for i in range(0,nSeg*2):
            if si_first:
                j = j+1
                si_first = not(si_first)
            else:
                jj=jj+1
                si_first = not(si_first)
            
            if ORDER: 
              xo.append((r+w/2)*math.cos(j*theta1+jj*theta2))
              yo.append((r+w/2)*math.sin(j*theta1+jj*theta2))
              xo.append((r-w/2)*math.cos(j*theta1+jj*theta2))
              yo.append((r-w/2)*math.sin(j*theta1+jj*theta2))
              ORDER = not(ORDER)
            else:
              xo.append((r-w/2)*math.cos(j*theta1+jj*theta2))
              yo.append((r-w/2)*math.sin(j*theta1+jj*theta2))
              xo.append((r+w/2)*math.cos(j*theta1+jj*theta2))
              yo.append((r+w/2)*math.sin(j*theta1+jj*theta2))
              ORDER = not(ORDER)    
          
            if len(xo) == 4:
              dpts=[pya.DPoint(xo[i], yo[i]) for i in range(len(xo))]
              dpolygon = DPolygon(dpts)
              element = Polygon.from_dpoly(dpolygon*(1.0/dbu))
              shapes(LayerSiN).insert(element)
              xo = []
              yo = [] 
        r = r - (w+gap) 
# Draw the Bus Waveguides 
#Draw Bus WG
      # to go back to the initial point for Multi-box drawing, ger the first adjusted r value
      r=self.r
      const = math.floor(2*pi*r/(s1+s2))
        #if doesn't divide evenly, replace r with best possible r
      if ((2*pi*r)%(s1+s2) != 0):
          r = const*(s1+s2)/(2*pi)
      
      #calulate ideal length of bus
      bus_length = busL
      #bus_length = self.cell.bbox().height()*dbu +pitch*2
      constant = math.ceil(bus_length/(s1+s2))
      if bus_length % (s1+s2) != 0:
        bus_length = constant*(s1+s2)
        
      for ii in range(0,int(row)): 
        xo = [(r+w/2+gap2+ii*(w+gap)),(r+w/2+gap2+w+ii*(w+gap)),(r+w/2+gap2+w+ii*(w+gap)),(r+w/2+gap2+ii*(w+gap))]
        yo = [0,0,s1,s1]
        dpts=[pya.DPoint(xo[i], yo[i]) for i in range(len(xo))]
        dpolygon = DPolygon(dpts)
        element = Polygon.from_dpoly(dpolygon*(1.0/dbu))
        shapes(LayerSiN).insert(element)
        
        # draw the bus waveguide     
        for i in range(0,int(math.ceil((constant)/2))):
          yu = [yo[j]+i*pitch for j in range(len(yo))]
          yd = [yo[j]-i*pitch for j in range(len(yo))]
            
          dpts=[pya.DPoint(xo[i], yu[i]) for i in range(len(xo))]
          dpolygon = DPolygon(dpts)
          element = Polygon.from_dpoly(dpolygon*(1.0/dbu))
          shapes(LayerSiN).insert(element)
          dpts=[pya.DPoint(xo[i], yd[i]) for i in range(len(xo))]
          dpolygon = DPolygon(dpts)
          element = Polygon.from_dpoly(dpolygon*(1.0/dbu))
          shapes(LayerSiN).insert(element) 
        
        
        # draw the tapers from waveguide to SWG
        xtu = [((r+w/2+gap2+ii*(w+gap))+(w-0.06)/2),((r+w/2+gap2+ii*(w+gap))+(w-0.06)/2+0.06),((r+w/2+gap2+ii*(w+gap))+w+gap/2),((r+w/2+gap2+ii*(w+gap))-gap/2)]
        ytu = [(yu[3]-taperL),(yu[3]-taperL),(yu[3]),(yu[3])]
        dpts=[pya.DPoint(xtu[i], ytu[i]) for i in range(len(xtu))]
        dpolygon = DPolygon(dpts)
        element = Polygon.from_dpoly(dpolygon*(1.0/dbu))
        shapes(LayerSiN).insert(element)
        
        ytd = [(yd[1]+taperL),(yd[1]+taperL),(yd[1]),(yd[1])]
        dpts=[pya.DPoint(xtu[i], ytd[i]) for i in range(len(xtu))]
        dpolygon = DPolygon(dpts)
        element = Polygon.from_dpoly(dpolygon*(1.0/dbu))
        shapes(LayerSiN).insert(element)
      
      # taper connections from multi-tapers to std WG
      xTu = [(r+w/2+gap2-gap/2),(r+w/2+gap2-gap/2+row*(gap+w)),(r+w/2+gap2+row*(gap+w)/2+0.25),(r+w/2+gap2+row*(gap+w)/2 - 0.25)]
      yTu = [(yu[3]),(yu[3]),(yu[3]+taperL),(yu[3]+taperL)]
      dpts=[pya.DPoint(xTu[i], yTu[i]) for i in range(len(xTu))]
      dpolygon = DPolygon(dpts)
      element = Polygon.from_dpoly(dpolygon*(1.0/dbu))
      shapes(LayerSiN).insert(element)
       
      yTd = [(yd[1]),(yd[1]),(yd[1]-taperL),(yd[1]-taperL)]
      dpts=[pya.DPoint(xTu[i], yTd[i]) for i in range(len(xTu))]
      dpolygon = DPolygon(dpts)
      element = Polygon.from_dpoly(dpolygon*(1.0/dbu))
      shapes(LayerSiN).insert(element) 
      
      # add std WG 
      if 2*r >= busL + 2*taperL:
        xstdu = [(r+w/2+gap2+row*(gap+w)/2  - 0.25),(r+w/2+gap2+row*(gap+w)/2  + 0.25),(r+w/2+gap2+row*(gap+w)/2 + 0.25),(r+w/2+gap2+row*(gap+w)/2 - 0.25)]
        ystdu = [(yu[3]+taperL),(yu[3]+taperL),(yu[3]+r-busL/2+1),(yu[3]+r-busL/2+1)]
        dpts=[pya.DPoint(xstdu[i], ystdu[i]) for i in range(len(xstdu))]
        dpolygon = DPolygon(dpts)
        element = Polygon.from_dpoly(dpolygon*(1.0/dbu))
        shapes(LayerSiN).insert(element)  
        
        ystdd = [(yd[1]-taperL),(yd[1]-taperL),(yd[1]-r+busL/2-1),(yd[1]-r+busL/2-1)]
        dpts=[pya.DPoint(xstdu[i], ystdd[i]) for i in range(len(xstdu))]
        dpolygon = DPolygon(dpts)
        element = Polygon.from_dpoly(dpolygon*(1.0/dbu))
        shapes(LayerSiN).insert(element) 
      else:
        xstdu = xTu 
        ystdu = yTu 
        ystdd = yTd 
        
      #DEV BOX
      half_l = (self.cell.bbox().width()-row*(gap+w)/dbu-gap2/dbu+(gap/2)/dbu)/2
      half_r = self.cell.bbox().width()-half_l
      dev = Box(-half_l,ystdd[-1]/dbu-1,half_r,ystdu[-1]/dbu+1)
      shapes(LayerDevRecN).insert(dev)
      dev_width = self.cell.bbox().width()/2
      
      if 2*r >= busL + 2*taperL:
        dev_up = yu[-1]/dbu +taperL/dbu +(yu[3]+r-busL/2 +1-(yu[3]+taperL))/dbu
        dev_down = yd[-4]/dbu - taperL/dbu - (yu[3]+r-busL/2 +1 -(yu[3]+taperL))/dbu
      else:
        dev_up = yu[-1]/dbu +taperL/dbu 
        dev_down = yd[-4]/dbu - taperL/dbu
         
      # Create the pins on the waveguides, as short paths:
      from SiEPIC._globals import PIN_LENGTH as pin_length
  
      w = to_itype(self.w,dbu)
      gap = to_itype(self.gap,dbu)
      bus_length = to_itype(bus_length/2,dbu)

      #Pin1
      #t = Trans(Trans.R0, dev_width-w/2,dev_up)
      t = Trans(Trans.R0, xstdu[-1]/dbu+251,dev_up+1)
      pin = Path([Point(0, -pin_length/2), Point(0, pin_length/2)], 500) #500 is width of std WG
      pin_t = pin.transformed(t)
      shapes(LayerPinRecN).insert(pin_t)
      text = Text ("opt1", t)
      shape = shapes(LayerPinRecN).insert(text)
      shape.text_size = 0.4/dbu
      if debug:
          print(dev_up)
          print(dev_up*dbu)
      #Pin2
      t = Trans(Trans.R0, xstdu[-1]/dbu+251,dev_down-1)
      pin = Path([Point(0, pin_length/2), Point(0, -pin_length/2)], 500)
      pin_t = pin.transformed(t)
      shapes(LayerPinRecN).insert(pin_t)
      text = Text ("opt2", t)
      shape = shapes(LayerPinRecN).insert(text)
      shape.text_size = 0.4/dbu



"""
Author:   Mustafa Hammood, 2022
          Mustafa@siepic.com; mustafa.hammood@dreamphotonics.com; mustafa@ece.ubc.ca
"""

import pya
from pya import *
from SiEPIC.utils import get_technology_by_name
from SiEPIC.utils.layout import make_pin, make_devrec_label


class contra_directional_coupler(pya.PCellDeclarationHelper):

    def __init__(self):
        # Important: initialize the super class
        super(contra_directional_coupler, self).__init__()
        TECHNOLOGY = get_technology_by_name('EBeam')

        # Button to launch separate window
        self.param("documentation", self.TypeCallback, "Open documentation in web browser")
        self.param("simulation", self.TypeCallback, "Launch simulation GUI")

        # declare the parameters
        self.param("number_of_periods", self.TypeInt, "Number of grating periods", default=1000)
        self.param("grating_period", self.TypeDouble,
                   "Grating period (microns)", default=0.316)
        self.param("gap", self.TypeDouble, "Gap (microns)", default=0.10)
        self.param("corrugation1_width", self.TypeDouble,
                   "Waveguide 1 Corrugration width (microns)", default=0.05)
        self.param("corrugation2_width", self.TypeDouble,
                   "Waveguide 2 Corrugration width (microns)", default=0.025)
        self.param("AR", self.TypeBoolean, "Anti-Reflection Design", default=True)
        self.param("sinusoidal", self.TypeBoolean,
                   "Grating Type (Rectangular=False, Sinusoidal=True)", default=False)
        self.param("wg1_width", self.TypeDouble, "Waveguide 1 width", default=0.56)
        self.param("wg2_width", self.TypeDouble, "Waveguide 2 width", default=0.44)
        self.param("sbend", self.TypeBoolean, "Include S-bends", default=1)
        self.param("sbend_r", self.TypeDouble, "S-bend radius (microns)", default=15)
        self.param("sbend_length", self.TypeDouble, "S-bend length (microns)", default=11)
        self.param("apodization_index", self.TypeDouble,
                   "Gaussian Apodization Index", default=10.0)
        self.param("port_w", self.TypeDouble, "Port Waveguide width", default=0.5)
        self.param("accuracy", self.TypeBoolean,
                   "Simulation Accuracy (on = high, off = fast)", default=True, hidden=True)
        self.param("layer", self.TypeLayer, "Waveguide Layer", default=TECHNOLOGY['Si'])
        self.param("rib", self.TypeBoolean, "Use rib (ridge) waveguides?", default=False)
        self.param("layer_rib", self.TypeLayer, "Rib Layer",
                   default=TECHNOLOGY['Si - 90 nm rib'])
        self.param("metal", self.TypeBoolean, "Use metal heaters?", default=False)
        self.param("heater_width", self.TypeDouble, "Metal heater width", default=3)
        self.param("metal_width", self.TypeDouble, "Metal contact width", default=5)
        self.param("l_heater", self.TypeLayer, "Heater Layer", default=TECHNOLOGY['M1_heater'])
        self.param("l_metal", self.TypeLayer, "Metal contact Layer", default=TECHNOLOGY['M2_router'])
        self.param("pinrecm", self.TypeLayer, "PinRecM Layer (metal)", default=TECHNOLOGY['PinRecM'])
        self.param("pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY['PinRec'])
        self.param("devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY['DevRec'])

    def callback(self, layout, name, states):
        ''' Callback for PCell, to launch documentation viewer
            https://www.klayout.de/doc/code/class_PCellDeclaration.html#method9
        '''
        if name == 'documentation':
            # Mustafa Hammood's repository for simulations of CDCs
            url = 'https://github.com/SiEPIC/contraDC'
            import webbrowser
            webbrowser.open_new(url)
        if name == 'simulation':
            print('contraDC simulation window')
            from SiEPIC.simulation.contraDC.klayout_gui import cdc_gui
            cdc_gui()
                        

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "contra_directional_coupler_%sN-%.1fnm period" % \
            (self.number_of_periods, self.grating_period*1000)

    def coerce_parameters_impl(self):
        pass

    def can_create_from_shape(self, layout, shape, layer):
        return False

    def produce_impl(self):
        import math
        try:
            from SiEPIC.utils.layout import layout_waveguide_sbend, layout_taper
        except:
            from siepic_tools.utils.layout import layout_waveguide_sbend, layout_taper

        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes

        LayerSi = self.layer
        LayerSiN = ly.layer(LayerSi)
        LayerRib = ly.layer(self.layer_rib)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)

        shapes_wg = pya.Region()
        shapes_rib = pya.Region()

        from SiEPIC.extend import to_itype

        # Draw the Bragg grating (bottom):
        box_width = int(round(self.grating_period/2/dbu))
        grating_period = int(round(self.grating_period/dbu))

        w = to_itype(self.wg1_width, dbu)
        GaussianIndex = self.apodization_index
        half_w = w/2
        half_corrugation_w = to_itype(self.corrugation1_width/2, dbu)

        y_offset_top = -w/2 - to_itype(self.gap/2, dbu)

        if self.AR:
            misalignment = grating_period/2
        else:
            misalignment = 0

        self.number_of_periods = int(self.number_of_periods)
        N = self.number_of_periods
        if self.sinusoidal:
            npoints_sin = 40
            for i in range(0, self.number_of_periods):
                x = (round((i * self.grating_period)/dbu))
                profileFunction = math.exp(-0.5*(2*GaussianIndex*(i-N/2)/(N))**2)
                profile = int(round(self.corrugation1_width/2/dbu))*profileFunction
                box1 = Box(x, y_offset_top, x + box_width, y_offset_top + half_w+profile)
                pts1 = [Point(x, y_offset_top)]
                pts3 = [Point(x + misalignment, y_offset_top)]

                for i1 in range(0, npoints_sin+1):
                    x1 = i1 * 2 * math.pi / npoints_sin
                    y1 = round(profile*math.sin(x1))
                    x1 = round(x1/2/math.pi*grating_period)
                    pts1.append(Point(x + x1, y_offset_top + half_w+y1))
                    pts3.append(Point(x + misalignment + x1, y_offset_top - half_w-y1))

                pts1.append(Point(x + grating_period, y_offset_top))
                pts3.append(Point(x + grating_period + misalignment, y_offset_top))

                shapes_wg += Polygon(pts1)
                shapes_wg += Polygon(pts3)

            length = x + grating_period + misalignment
            if misalignment > 0:
                # extra piece at the end:
                box2 = Box(x + grating_period, y_offset_top, length, y_offset_top + half_w)
                # extra piece at the beginning:
                box3 = Box(0, y_offset_top, misalignment, y_offset_top - half_w)

                shapes_wg += box2
                shapes_wg += box3

        else:
            for i in range(0, self.number_of_periods):
                x = int(round((i * self.grating_period)/dbu))

                profileFunction = math.exp(-0.5*(2*GaussianIndex*(i-N/2)/(N))**2)
                profile = int(round(self.corrugation1_width/2/dbu))*profileFunction

                box1 = Box(x, y_offset_top, x + box_width, y_offset_top +
                           to_itype(half_w+profile, dbu*1000))
                box2 = Box(x + box_width, y_offset_top, x + grating_period,
                           y_offset_top + to_itype(half_w-profile, dbu*1000))
                box3 = Box(x + misalignment, y_offset_top, x + box_width +
                           misalignment, y_offset_top + to_itype(-half_w-profile, dbu*1000))
                box4 = Box(x + box_width + misalignment, y_offset_top, x + grating_period +
                           misalignment, y_offset_top + to_itype(-half_w+profile, dbu*1000))

                shapes_wg += box1
                shapes_wg += box2
                shapes_wg += box3
                shapes_wg += box4
            length = x + grating_period + misalignment
            if misalignment > 0:
                # extra piece at the end:
                box2 = Box(x + grating_period, y_offset_top, length, y_offset_top + half_w)
                shapes_wg += box2
                # extra piece at the beginning:
                box3 = Box(0, y_offset_top, misalignment, y_offset_top - half_w)
                shapes_wg += box3

        vertical_offset = int(round(self.wg2_width/2/dbu))+int(round(self.gap/2/dbu))

        if misalignment > 0:
            t = Trans(Trans.R0, 0, vertical_offset)
        else:
            t = Trans(Trans.R0, 0, vertical_offset)

        # Draw the Bragg grating (top):
        box_width = int(round(self.grating_period/2/dbu))
        grating_period = int(round(self.grating_period/dbu))
        w = to_itype(self.wg2_width, dbu)
        half_w = w/2
        half_corrugation_w = int(round(self.corrugation2_width/2/dbu))

        N = self.number_of_periods
        if self.sinusoidal:
            npoints_sin = 40
            for i in range(0, self.number_of_periods):
                x = (round((i * self.grating_period)/dbu))
                profileFunction = math.exp(-0.5*(2*GaussianIndex*(i-N/2)/(N))**2)
                profile = int(round(self.corrugation2_width/2/dbu))*profileFunction
                box1 = Box(x, 0, x + box_width, -half_w+profile).transformed(t)
                pts1 = [Point(x, 0)]
                pts3 = [Point(x + misalignment, 0)]
                for i1 in range(0, npoints_sin+1):
                    x1 = i1 * 2 * math.pi / npoints_sin
                    y1 = round(profile*math.sin(x1))
                    x1 = round(x1/2/math.pi*grating_period)
#          print("x: %s, y: %s" % (x1,y1))
                    pts1.append(Point(x + x1, -half_w-y1))
                    pts3.append(Point(x + misalignment + x1, +half_w+y1))
                pts1.append(Point(x + grating_period, 0))
                pts3.append(Point(x + grating_period + misalignment, 0))

                shapes_wg += Polygon(pts1).transformed(t)
                shapes_wg += Polygon(pts3).transformed(t)
            length = x + grating_period + misalignment
            if misalignment > 0:
                # extra piece at the end:
                box2 = Box(x + grating_period, 0, length, -half_w).transformed(t)
                shapes_wg += box2
                # extra piece at the beginning:
                box3 = Box(0, 0, misalignment, half_w).transformed(t)
                shapes_wg += box3

        else:
            for i in range(0, self.number_of_periods):
                x = int(round((i * self.grating_period)/dbu))
                profileFunction = math.exp(-0.5*(2*GaussianIndex*(i-N/2)/(N))**2)
                profile = int(round(self.corrugation2_width/2/dbu))*profileFunction
                box1 = Box(x, 0, x + box_width, -half_w-profile).transformed(t)
                box2 = Box(x + box_width, 0, x + grating_period, -
                           half_w+profile).transformed(t)
                box3 = Box(x + misalignment, 0, x + box_width +
                           misalignment, half_w+profile).transformed(t)
                box4 = Box(x + box_width + misalignment, 0, x + grating_period +
                           misalignment, half_w-profile).transformed(t)

                shapes_wg += box1
                shapes_wg += box2
                shapes_wg += box3
                shapes_wg += box4

            length = x + grating_period + misalignment
            if misalignment > 0:
                # extra piece at the end:
                box2 = Box(x + grating_period, 0, length, -half_w).transformed(t)
                shapes_wg += box2
                # extra piece at the beginning:
                box3 = Box(0, 0, misalignment, half_w).transformed(t)
                shapes_wg += box3

        # Create the pins on the waveguides, as short paths:
        from SiEPIC._globals import PIN_LENGTH as pin_length

        w1 = to_itype(self.wg1_width, dbu)
        w2 = to_itype(self.wg2_width, dbu)
        if self.sbend:

            port_w = to_itype(self.port_w, dbu)
            sbend_r = to_itype(self.sbend_r, dbu)
            sbend_length = to_itype(self.sbend_length, dbu)
            sbend_offset = 4*port_w + port_w - (w1+w2)/2 - int(round(self.gap/dbu))
            taper_length = 50*max(abs(w1-port_w), abs(w2-port_w))

            t = Trans(Trans.R180, 0, y_offset_top)
            shapes_wg += layout_waveguide_sbend(self.cell, LayerSiN,
                                                t, w1, sbend_r, sbend_offset, sbend_length, insert=False, dbu=dbu)
            t = Trans(Trans.R0, -sbend_length-taper_length, y_offset_top-sbend_offset)
            shapes_wg += layout_taper(self.cell, LayerSiN, t,  port_w,
                                      w1, taper_length, insert=False)
            x = -sbend_length-taper_length
            y = y_offset_top-sbend_offset
            make_pin(self.cell, "opt1", [x, y], port_w, LayerPinRecN, 180)

            x = -taper_length-sbend_length
            y = vertical_offset
            make_pin(self.cell, "opt2", [x, y], port_w, LayerPinRecN, 180)

            x = length+sbend_length+taper_length
            y = y_offset_top-sbend_offset
            # i don't know why int(x) works but x doesnt
            # future debugger, email mustafa hammood :)
            make_pin(self.cell, "opt3", [int(x), y], port_w, LayerPinRecN, 0)

            x = length+taper_length+sbend_length
            y = vertical_offset
            make_pin(self.cell, "opt4", [int(x), y], port_w, LayerPinRecN, 0)

            t = Trans(Trans.R180, 0, vertical_offset)
            shapes_wg += layout_taper(self.cell, LayerSiN, t,  w2,
                                      w2, sbend_length/2, insert=False)
            t = Trans(Trans.R180, -sbend_length/2, vertical_offset)
            shapes_wg += layout_taper(self.cell, LayerSiN, t,  w2,
                                      port_w, taper_length+sbend_length/2, insert=False)


            t = Trans(Trans.R0, length, y_offset_top)
            shapes_wg += layout_waveguide_sbend(self.cell, LayerSiN, t,
                                                w1, sbend_r, -sbend_offset, sbend_length, insert=False, dbu=dbu)
            t = Trans(Trans.R0, length+sbend_length, y_offset_top-sbend_offset)
            shapes_wg += layout_taper(self.cell, LayerSiN, t, w1,
                                      port_w, taper_length, insert=False)


            t = Trans(Trans.R0, length, vertical_offset)
            shapes_wg += layout_taper(self.cell, LayerSiN, t,  w2,
                                      w2, sbend_length/2, insert=False)
            t = Trans(Trans.R0, length+sbend_length/2, vertical_offset)
            shapes_wg += layout_taper(self.cell, LayerSiN, t,  w2,
                                      port_w, taper_length+sbend_length/2, insert=False)

        else:
            x = 0
            y = y_offset_top
            make_pin(self.cell, "opt1", [x, y], w1, LayerPinRecN, 180)

            y = vertical_offset
            make_pin(self.cell, "opt2", [x, y], w2, LayerPinRecN, 180)

            x = length
            y = y_offset_top
            make_pin(self.cell, "opt3", [int(x), y], w1, LayerPinRecN, 0)

            x = length
            y = vertical_offset
            make_pin(self.cell, "opt4", [int(x), y], w2, LayerPinRecN, 0)

        # Compact model information. Create the devrec library labels
        make_devrec_label(self.cell, "EBeam", "contra_directional_coupler",
                          ly.layer(self.devrec))

        text = Text('Spice_param:number_of_periods=%s grating_period=%.4fu wg1_width=%.3fu wg2_width=%.3fu corrugation1_width=%.3fu corrugation2_width=%.3fu gap=%.3fu apodization_index=%.3f AR=%s sinusoidal=%s accuracy=%s' %
                    (self.number_of_periods, self.grating_period, self.wg1_width, self.wg2_width, self.corrugation1_width, self.corrugation2_width, self.gap, self.apodization_index, int(self.AR), int(self.sinusoidal), int(self.accuracy)), t)
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = 0.1/dbu
        
        if self.metal:
            # add heater element
            heater_width = to_itype(self.heater_width, dbu)
            box = pya.Box(pya.Point(0, heater_width/2),
                          pya.Point(length, -heater_width/2))
            shapes(ly.layer(self.l_heater)).insert(box)

            # add metal contact element (left and right)
            metal_width = to_itype(self.metal_width, dbu)
            box = pya.Box(pya.Point(0, heater_width/2),
                          pya.Point(metal_width, -heater_width/2))
            shapes(ly.layer(self.l_metal)).insert(box)
            box = pya.Box(pya.Point(length, heater_width/2),
                          pya.Point(length-metal_width, -heater_width/2))
            shapes(ly.layer(self.l_metal)).insert(box)

            # add metal contact pins
            x = metal_width/2 * dbu
            y = -heater_width/2 * dbu
            make_pin(self.cell, "elec1", [x, y], metal_width, ly.layer(self.pinrecm), 270)
            x = (length-metal_width/2) * dbu
            make_pin(self.cell, "elec2", [x, y], metal_width, ly.layer(self.pinrecm), 270)
            

        # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
        if self.sbend:
            box = pya.Box(pya.Point(-taper_length-sbend_length, vertical_offset+3*port_w),
                          pya.Point(length+taper_length+sbend_length, y_offset_top-sbend_offset-3*port_w))
        else:
            box = pya.Box(pya.Point(0, vertical_offset+3*(w1+w2)/2),
                          pya.Point(length, y_offset_top-3*(w1+w2)/2))
        DevRecBox = box
        shapes(LayerDevRecN).insert(box)

        # Draw the waveguide layer
        if self.rib == False:
            shapes(LayerSiN).insert(shapes_wg)
        else:  # turn shape into a rib waveguide
            region_devrec = Region(DevRecBox)
            region_devrec2 = Region(DevRecBox).size(1500)
            shapes_rib += shapes_wg
            shapeRib = shapes_rib.size(1400) - shapes_wg - (region_devrec2-region_devrec)

            shapes(LayerRib).insert(shapeRib)
            shapes(LayerRib).insert(shapes_wg)
            shapes(LayerSiN).insert(shapes_wg)
